// Modified by Gang Wang
// Modified by Andrew Fiore

#include "Helper.cuh"

//! Shared memory array for partial sum of dot product kernel
extern __shared__ Scalar partial_sum[];
//! Shared memory array for rigid body calculation
extern __shared__ Scalar3 shared_rigid_vel[];

/*!
	Dot product helper function: First step
	d_a .* d_b -> d_c -> Partial sum
	BlockDim of this kernel should be 2^n, which is 512. (Based on HOOMD ComputeThermoGPU class)
*/
__global__
void gpu_stokes_DotStepOne_kernel(Scalar4 *d_a, Scalar4 *d_b, Scalar *dot_sum, unsigned int group_size, unsigned int *d_group_members){
	int group_idx = blockDim.x * blockIdx.x + threadIdx.x;
	Scalar temp;

	if (group_idx < group_size) {

		unsigned int idx = d_group_members[group_idx];
		Scalar4 a4 = d_a[idx];
		Scalar4 b4 = d_b[idx];
		Scalar3 a = make_scalar3(a4.x, a4.y, a4.z);
		Scalar3 b = make_scalar3(b4.x, b4.y, b4.z);

		temp = dot(a,b); // Partial sum, each thread, shared memory

	}
	else {
		temp = 0;
	}

	partial_sum[threadIdx.x] = temp;

	__syncthreads();

	int offs = blockDim.x >> 1;

	while (offs > 0)
        {
        	if (threadIdx.x < offs)
            	{
            		partial_sum[threadIdx.x] += partial_sum[threadIdx.x + offs];
            	}
        	offs >>= 1;
        	__syncthreads();
        }

	if (threadIdx.x == 0){
		dot_sum[blockIdx.x] = partial_sum[0];
	}
}

/*!
	Dot product helper function: Second step
	Partial sum -> Final sum
	Only one block will be launched for this step
*/
__global__
void gpu_stokes_DotStepTwo_kernel(Scalar *dot_sum, unsigned int num_partial_sums){

	partial_sum[threadIdx.x] = 0.0;
	__syncthreads();
	for (unsigned int start = 0; start < num_partial_sums; start += blockDim.x)
       	{
        	if (start + threadIdx.x < num_partial_sums)
            	{
            		partial_sum[threadIdx.x] += dot_sum[start + threadIdx.x];
            	}
	}

	int offs = blockDim.x >> 1;
	while (offs > 0)
       	{
		__syncthreads();
            	if (threadIdx.x < offs)
                {
                	partial_sum[threadIdx.x] += partial_sum[threadIdx.x + offs];
                }
            	offs >>= 1;

        }
	__syncthreads();
        if (threadIdx.x == 0)
	{
            	dot_sum[0] = partial_sum[0]; // Save the dot product to the first element of dot_sum array
	}

}


/*!
	Scalar multiplication
	C = a*A
	C can be A, so that A will be overwritten
	\param coeff_a a
*/
__global__
void gpu_stokesrigid_ScalarMtpl_kernel(Scalar4 *d_a, Scalar4 *d_c, Scalar coeff_a, unsigned int group_size, unsigned int *d_group_members){
	int group_idx = blockDim.x * blockIdx.x + threadIdx.x;
	if (group_idx < group_size) {
		unsigned int idx = d_group_members[group_idx];
		Scalar4 A4 = d_a[idx];
		Scalar3 A = make_scalar3(A4.x, A4.y, A4.z);
		A = coeff_a * A;
		d_c[idx] = make_scalar4(A.x, A.y, A.z, 0.0);
	}
}


/*!
	Linear combination helper function
	C = a*A + b*B
	C can be A or B, so that A or B will be overwritten
	\param coeff_a a
	\param coeff_b b
*/
__global__
void gpu_stokes_LinearCombination_kernel(Scalar4 *d_a, Scalar4 *d_b, Scalar4 *d_c, Scalar coeff_a, Scalar coeff_b, unsigned int group_size, unsigned int *d_group_members){
	int group_idx = blockDim.x * blockIdx.x + threadIdx.x;
	if (group_idx < group_size) {
		unsigned int idx = d_group_members[group_idx];
		Scalar4 A4 = d_a[idx];
		Scalar4 B4 = d_b[idx];
		Scalar3 A = make_scalar3(A4.x, A4.y, A4.z);
		Scalar3 B = make_scalar3(B4.x, B4.y, B4.z);
		A = coeff_a * A + coeff_b * B;
		d_c[idx] = make_scalar4(A.x, A.y, A.z, 0.0);
	}
}


/*!
	Linear combination helper function 2
	D = a*A + b*B + c*C
	D can be A or B or C, so that A or B or C will be overwritten
	\param coeff_a a
	\param coeff_b b
	\param coeff_c c
*/
__global__
void gpu_stokes_LinearCombination2_kernel(Scalar4 *d_a, Scalar4 *d_b, Scalar4 *d_c, Scalar4 *d_d, Scalar coeff_a, Scalar coeff_b, Scalar coeff_c, unsigned int group_size, unsigned int *d_group_members){
	int group_idx = blockDim.x * blockIdx.x + threadIdx.x;
	if (group_idx < group_size) {
		unsigned int idx = d_group_members[group_idx];
		Scalar4 A4 = d_a[idx];
		Scalar4 B4 = d_b[idx];
		Scalar4 C4 = d_c[idx];
		Scalar3 A = make_scalar3(A4.x, A4.y, A4.z);
		Scalar3 B = make_scalar3(B4.x, B4.y, B4.z);
		Scalar3 C = make_scalar3(C4.x, C4.y, C4.z);
		A = coeff_a * A + coeff_b * B + coeff_c * C;
		d_d[idx] = make_scalar4(A.x, A.y, A.z, 0.0);
	}
}


/*!
	Set vector to a same value helper function

	d_a = a
*/
__global__
void gpu_stokesrigid_SetValue_kernel(Scalar4 *d_a, Scalar3 a, unsigned int group_size, unsigned int *d_group_members){
	int group_idx = blockDim.x * blockIdx.x + threadIdx.x;
	if (group_idx < group_size) {
		unsigned int idx = d_group_members[group_idx];

		d_a[idx] = make_scalar4(a.x, a.y, a.z, 0.0);
	}
}

/*!
	Set vector equal to another vector helper function

	d_a := d_b
*/
__global__
void gpu_stokesrigid_SetEqual_kernel(Scalar4 *d_a, Scalar4 *d_b, unsigned int group_size, unsigned int *d_group_members){
	int group_idx = blockDim.x * blockIdx.x + threadIdx.x;
	if (group_idx < group_size) {
		unsigned int idx = d_group_members[group_idx];
		Scalar4 B = d_b[idx];
		d_a[idx] = make_scalar4(B.x, B.y, B.z, 0.0);
	}
}


//! kernel function to add forces for each particle to each rigid body
// Sigma matrix defined in Swan et al., Physics of fluids, 2011
// d_net_force -> d_rigidforce & d_rigidtorque
__global__
void gpu_stokesrigid_rigidsigma_kernel(Scalar4 *d_net_force,
				Scalar4 *d_rigidforce,
				Scalar4 *d_rigidtorque,
				unsigned int *d_rigidindex,
				Scalar4 *d_rigidorientation,
				unsigned int n_bodies,
				unsigned int nmax,
				unsigned int *d_rigidparticleindices,
				Scalar4 *d_rigidparticlepos
				)
{
	unsigned int body_group_idx = blockIdx.x * blockDim.y + threadIdx.y;
	unsigned int group_idx = threadIdx.x;
	unsigned int body_idx;
	unsigned int idx;

	Scalar4 forcei, posi; // Force of single particle; local position (COM frame) of single particle
	Scalar3 ri; // local position (space frame, relative to COM) of single particle
	Scalar3 temp_force_body = make_scalar3(0.0,0.0,0.0); // Temp vector for force reading
	Scalar3 temp_torque_body = make_scalar3(0.0,0.0,0.0); // Temp vector for torque calculating
	Scalar4 ex_space, ey_space, ez_space;

	if (body_group_idx < n_bodies){
		body_idx = d_rigidindex[body_group_idx];
		Scalar4 quat = d_rigidorientation[body_idx];

		exyzFromQuaternion(quat, ex_space, ey_space, ez_space);

		// read particle force
		for (unsigned int start = 0; start < nmax; start += blockDim.x){
			if (group_idx + start < nmax){
				unsigned int local_idx = body_idx * nmax + group_idx + start;
				idx = d_rigidparticleindices[local_idx];
				if (idx != INVALID_INDEX){
					forcei = d_net_force[idx];
					temp_force_body += make_scalar3(forcei.x, forcei.y, forcei.z);
					posi = d_rigidparticlepos[local_idx];
					ri.x = ex_space.x * posi.x + ey_space.x * posi.y + ez_space.x * posi.z;
                			ri.y = ex_space.y * posi.x + ey_space.y * posi.y + ez_space.y * posi.z;
                			ri.z = ex_space.z * posi.x + ey_space.z * posi.y + ez_space.z * posi.z;
					temp_torque_body += make_scalar3(ri.y * forcei.z - ri.z * forcei.y, ri.z * forcei.x - ri.x * forcei.z ,ri.x * forcei.y - ri.y * forcei.x );
				}
			}
		}
	}

	// Save them into shared memory
	shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y] = temp_force_body;
	shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y + blockDim.x * blockDim.y] = temp_torque_body;

	// Reduction with each rigid body
	// Note: This seems to be not the most efficient way of doing reduction. (Thread divergence)
	// A better practice should be switching threadIdx.x and threadIdx.y, which will be implemented later, since this is not the rate-limiting step..
	int offs = blockDim.x >> 1;
	while (offs > 0)
       	{
		__syncthreads();
            	if (threadIdx.x < offs)
                {
                	shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y] += shared_rigid_vel[threadIdx.x + offs + blockDim.x * threadIdx.y];
			shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y + blockDim.x * blockDim.y] += shared_rigid_vel[threadIdx.x + offs + blockDim.x * threadIdx.y + blockDim.x * blockDim.y];
                }
            	offs >>= 1;
        }
	__syncthreads();
        if (threadIdx.x == 0)
	{
		if (body_group_idx < n_bodies){
			temp_force_body = shared_rigid_vel[blockDim.x * threadIdx.y];
			temp_torque_body = shared_rigid_vel[blockDim.x * threadIdx.y + blockDim.x * blockDim.y];
			d_rigidforce[body_idx] = make_scalar4(temp_force_body.x,temp_force_body.y,temp_force_body.z,0.0);
			d_rigidtorque[body_idx] = make_scalar4(temp_torque_body.x, temp_torque_body.y, temp_torque_body.z, 0.0);

		}
	}
	// Save force and torque to global memory
}


//! kernel function to get velocity for each particle from each rigid body
// Sigma matrix transpose defined in Swan et al., Physics of fluids, 2011
// d_rigidvel & d_rigidangvel -> d_vel
__global__
void gpu_stokesrigid_rigidsigmatrans_kernel(Scalar4 *d_vel,
				Scalar4 *d_rigidvel,
				Scalar4 *d_rigidangvel,
				unsigned int *d_rigidindex,
				Scalar4 *d_rigidorientation,
				unsigned int n_bodies,
				unsigned int nmax,
				unsigned int *d_rigidparticleindices,
				Scalar4 *d_rigidparticlepos
				)
{
	unsigned int body_group_idx = blockIdx.x * blockDim.y + threadIdx.y;
	unsigned int group_idx = threadIdx.x;
	unsigned int body_idx;
	unsigned int idx;

	if (body_group_idx < n_bodies){
		body_idx = d_rigidindex[body_group_idx];

		Scalar4 quat = d_rigidorientation[body_idx];
		Scalar4 vel_body = d_rigidvel[body_idx];
		Scalar4 angvel_body = d_rigidangvel[body_idx];
		Scalar4 ex_space, ey_space, ez_space;

		exyzFromQuaternion(quat, ex_space, ey_space, ez_space);

		// read particle position & write velocity
		for (unsigned int start = 0; start < nmax; start += blockDim.x){
			if (group_idx + start < nmax){
				unsigned int local_idx = body_idx * nmax + group_idx + start;
				idx = d_rigidparticleindices[local_idx];
				if (idx != INVALID_INDEX){

					Scalar4 posi = d_rigidparticlepos[local_idx]; // Local position (COM frame) of single particle
					Scalar3 ri; // Local position (space frame, relative to COM) of single particle

					ri.x = ex_space.x * posi.x + ey_space.x * posi.y + ez_space.x * posi.z;
                			ri.y = ex_space.y * posi.x + ey_space.y * posi.y + ez_space.y * posi.z;
                			ri.z = ex_space.z * posi.x + ey_space.z * posi.y + ez_space.z * posi.z;

					Scalar3 veli = make_scalar3(vel_body.x, vel_body.y, vel_body.z);
					veli += make_scalar3(angvel_body.y * ri.z - angvel_body.z * ri.y, angvel_body.z * ri.x - angvel_body.x * ri.z ,angvel_body.x * ri.y - angvel_body.y * ri.x);
					d_vel[idx] = make_scalar4(veli.x, veli.y, veli.z, 0.0);
				}
			}
		}
	}
}


//! kernel function to add forces for each particle to each rigid body
// Sigma matrix defined in Swan et al., Physics of fluids, 2011
// d_net_force -> d_rigidforce & d_rigidtorque
__global__
void gpu_stokesrigid_rigidsigmaStress_kernel(Scalar4 *d_net_force,
				Scalar3 *d_vecSii,
			        Scalar3 *d_vecSij,
				unsigned int *d_rigidindex,
				Scalar4 *d_rigidorientation,
				unsigned int n_bodies,
				unsigned int nmax,
				unsigned int *d_rigidparticleindices,
				Scalar4 *d_rigidparticlepos
				)
{
	unsigned int body_group_idx = blockIdx.x * blockDim.y + threadIdx.y;
	unsigned int group_idx = threadIdx.x;
	unsigned int body_idx;
	unsigned int idx;

	Scalar4 forcei, posi; // Force of single particle; local position (COM frame) of single particle
	Scalar3 ri; // local position (space frame, relative to COM) of single particle
	Scalar3 temp_Sii = make_scalar3(0.0,0.0,0.0); // Temp vector for stress calculating
	Scalar3 temp_Sij = make_scalar3(0.0,0.0,0.0); // Temp vector for stress calculating
	Scalar4 ex_space, ey_space, ez_space;

	if (body_group_idx < n_bodies){
		body_idx = d_rigidindex[body_group_idx];
		Scalar4 quat = d_rigidorientation[body_idx];

		exyzFromQuaternion(quat, ex_space, ey_space, ez_space);

		// read particle force
		for (unsigned int start = 0; start < nmax; start += blockDim.x){
			if (group_idx + start < nmax){
				unsigned int local_idx = body_idx * nmax + group_idx + start;
				idx = d_rigidparticleindices[local_idx];
				if (idx != INVALID_INDEX){
					forcei = d_net_force[idx];
					posi = d_rigidparticlepos[local_idx];
					ri.x = ex_space.x * posi.x + ey_space.x * posi.y + ez_space.x * posi.z;
                			ri.y = ex_space.y * posi.x + ey_space.y * posi.y + ez_space.y * posi.z;
                			ri.z = ex_space.z * posi.x + ey_space.z * posi.y + ez_space.z * posi.z;
					temp_Sii += make_scalar3(ri.x * forcei.x, ri.y * forcei.y, ri.z * forcei.z);
					temp_Sij += (0.5 * make_scalar3(ri.y * forcei.x + ri.x * forcei.y, ri.z * forcei.x + ri.x * forcei.z, ri.z * forcei.y + ri.y * forcei.z));
				}
			}
		}
	}

	// Save them into shared memory
	shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y] = temp_Sii;
	shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y + blockDim.x * blockDim.y] = temp_Sij;

	// Reduction with each rigid body
	int offs = blockDim.x >> 1;
	while (offs > 0)
       	{
		__syncthreads();
            	if (threadIdx.x < offs)
                {
                	shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y] += shared_rigid_vel[threadIdx.x + offs + blockDim.x * threadIdx.y];
			shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y + blockDim.x * blockDim.y] += shared_rigid_vel[threadIdx.x + offs + blockDim.x * threadIdx.y + blockDim.x * blockDim.y];
                }
            	offs >>= 1;
        }
	__syncthreads();
        if (threadIdx.x == 0)
	{
		if (body_group_idx < n_bodies){
			d_vecSii[body_idx] = shared_rigid_vel[blockDim.x * threadIdx.y];
			d_vecSij[body_idx] = shared_rigid_vel[blockDim.x * threadIdx.y + blockDim.x * blockDim.y];
		}
	}
	// Save the stress to global memory
}


/*!
	Vector sum helper function: First step
	d_vector -> Partial sum
	BlockDim of this kernel should be 2^n, which is 512. (Based on HOOMD ComputeThermoGPU class)
*/
__global__
void gpu_stokesrigid_VectorSumStepOne_kernel(Scalar3 *d_a, Scalar3 *vector_sum, unsigned int group_size, unsigned int *d_group_members){
	int group_idx = blockDim.x * blockIdx.x + threadIdx.x;
	Scalar3 temp;

	if (group_idx < group_size) {
		unsigned int idx = d_group_members[group_idx];
		temp = d_a[idx];
	}
	else {
		temp = make_scalar3(0.0,0.0,0.0);
	}

	shared_rigid_vel[threadIdx.x] = temp;

	__syncthreads();

	int offs = blockDim.x >> 1;

	while (offs > 0)
        {
        	if (threadIdx.x < offs)
            	{
            		shared_rigid_vel[threadIdx.x] += shared_rigid_vel[threadIdx.x + offs];
            	}
        	offs >>= 1;
        	__syncthreads();
        }

	if (threadIdx.x == 0){
		vector_sum[blockIdx.x] = shared_rigid_vel[0];
	}
}


// For sedimentation simulation: start
/*!
	Vector sum helper function: First step
	d_vector -> Partial sum
	BlockDim of this kernel should be 2^n, which is 512. (Based on HOOMD ComputeThermoGPU class)
	For Scalar4 type.
*/
__global__
void gpu_stokesrigid_VectorSumStepOne4_kernel(Scalar4 *d_a, Scalar3 *vector_sum, unsigned int group_size, unsigned int *d_group_members){
	int group_idx = blockDim.x * blockIdx.x + threadIdx.x;
	Scalar3 temp;

	if (group_idx < group_size) {
		unsigned int idx = d_group_members[group_idx];
		Scalar4 temp4 = d_a[idx];
		temp = make_scalar3(temp4.x, temp4.y, temp4.z);
	}
	else {
		temp = make_scalar3(0.0,0.0,0.0);
	}

	shared_rigid_vel[threadIdx.x] = temp;

	__syncthreads();

	int offs = blockDim.x >> 1;

	while (offs > 0)
        {
        	if (threadIdx.x < offs)
            	{
            		shared_rigid_vel[threadIdx.x] += shared_rigid_vel[threadIdx.x + offs];
            	}
        	offs >>= 1;
        	__syncthreads();
        }

	if (threadIdx.x == 0){
		vector_sum[blockIdx.x] = shared_rigid_vel[0];
	}
}
// For sedimentation simulation: end


/*!
	Vector sum helper function: Second step
	Partial sum -> Final sum: sum over all the particles (Sum_vector)
	Only one block will be launched for this step
*/
__global__
void gpu_stokesrigid_VectorSumStepTwo_kernel(Scalar3 *vector_sum, unsigned int num_partial_sums){

	shared_rigid_vel[threadIdx.x] = make_scalar3(0.0,0.0,0.0);
	__syncthreads();
	for (unsigned int start = 0; start < num_partial_sums; start += blockDim.x)
       	{
        	if (start + threadIdx.x < num_partial_sums)
            	{
            		shared_rigid_vel[threadIdx.x] += vector_sum[start + threadIdx.x];
            	}
	}

	int offs = blockDim.x >> 1;
	while (offs > 0)
       	{
		__syncthreads();
            	if (threadIdx.x < offs)
                {
                	shared_rigid_vel[threadIdx.x] += shared_rigid_vel[threadIdx.x + offs];
                }
            	offs >>= 1;

        }
	__syncthreads();
        if (threadIdx.x == 0)
	{
            	vector_sum[0] = shared_rigid_vel[0];
	}

}


//! Zero out the force grid
/*!
	\param grid the grid going to be zero out
   	\param NxNyNz dimension of the grid
*/
__global__
void gpu_stokes_ZeroGrid_kernel(CUFFTCOMPLEX *grid, unsigned int NxNyNz) {

  unsigned int tid = blockDim.x * blockIdx.x + threadIdx.x;

  if ( tid < NxNyNz ) {

  	grid[tid] = make_scalar2( 0.0, 0.0 );

  }
}


/*!

	Perform matrix-vector multiply needed for the Lanczos contribution to the Brownian velocity

	\param d_A 		matrix, N x m
	\param d_x		multiplying vector, m x 1
	\param d_b		result vector, A*x, m x 1
	\param group_size	number of particles
	\param m		number of iterations ( number of columns of A, length of x )

*/
__global__
void gpu_stokes_MatVecMultiply_kernel(Scalar4 *d_A, Scalar *d_x, Scalar4 *d_b, unsigned int group_size, int m){
	int idx = blockDim.x * blockIdx.x + threadIdx.x;
	if (idx < group_size) {

		Scalar3 tempprod = make_scalar3( 0.0, 0.0, 0.0 );

		for ( int ii = 0; ii < m; ++ii ){

		    Scalar4 matidx = d_A[ idx + ii*group_size ];

		    Scalar xcurr = d_x[ii];

		    tempprod.x = tempprod.x + matidx.x * xcurr;
		    tempprod.y = tempprod.y + matidx.y * xcurr;
		    tempprod.z = tempprod.z + matidx.z * xcurr;

		}

		d_b[idx] = make_scalar4( tempprod.x, tempprod.y, tempprod.z, d_A[idx].w );

	}
}


/*!
	Add two grid vectors
	C = A + B

	\param d_a              input vector, A
	\param d_b              input vector, B
	\param d_c              output vector, C
	\param N                length of vectors
*/
__global__
void gpu_stokes_AddGrids_kernel(CUFFTCOMPLEX *d_a, CUFFTCOMPLEX *d_b, CUFFTCOMPLEX *d_c, unsigned int NxNyNz){
	int tidx = blockDim.x * blockIdx.x + threadIdx.x;
	if ( tidx < NxNyNz) {
		unsigned int idx = tidx;
		CUFFTCOMPLEX A = d_a[idx];
		CUFFTCOMPLEX B = d_b[idx];
		d_c[idx] = make_scalar2(A.x+B.x, A.y+B.y);
	}
}


/*!
	Add scale the grid
	A = s * A

	\param d_a   input vector, A
	\param s     scale factor
	\param N     length of vectors
*/
__global__
void gpu_stokes_ScaleGrid_kernel(CUFFTCOMPLEX *d_a, Scalar s, unsigned int NxNyNz){
	int tidx = blockDim.x * blockIdx.x + threadIdx.x;
	if ( tidx < NxNyNz) {
		unsigned int idx = tidx;
		CUFFTCOMPLEX A = d_a[idx];
		d_a[idx] = make_scalar2(s*A.x, s*A.y);
	}
}

/*!
  Kernel function to calculate position of each grid in reciprocal space: gridk
  */
__global__
void gpu_stokes_SetGridk_kernel(Scalar4 *gridk,
                                int Nx,
                                int Ny,
                                int Nz,
                                unsigned int NxNyNz,
                                BoxDim box,
                                Scalar xi,
				Scalar eta)
{
        int tid = blockDim.x * blockIdx.x + threadIdx.x;

        if ( tid < NxNyNz ) {

                int i = tid / (Ny*Nz);
                int j = (tid - i * Ny * Nz) / Nz;
                int k = tid % Nz;

                Scalar3 L = box.getL();
                Scalar xy = box.getTiltFactorXY();
                Scalar4 gridk_value;

                gridk_value.x = (i < (Nx+1) / 2) ? i : i - Nx;
                gridk_value.y = ( ((j < (Ny+1) / 2) ? j : j - Ny) - xy * gridk_value.x * L.y / L.x ) / L.y; // Fixed by Zsigi 2015
                gridk_value.x = gridk_value.x / L.x;
                gridk_value.z = ((k < (Nz+1) / 2) ? k : k - Nz) / L.z;

		gridk_value.x *= 2.0*3.1416926536;
		gridk_value.y *= 2.0*3.1416926536;
		gridk_value.z *= 2.0*3.1416926536;

                Scalar k2 = gridk_value.x*gridk_value.x + gridk_value.y*gridk_value.y + gridk_value.z*gridk_value.z;
		Scalar xisq = xi * xi;

		// Scaling factor used in wave space sum
		if (i == 0 && j == 0 && k == 0){
			gridk_value.w = 0.0;
		}
		else{
			// Have to divide by Nx*Ny*Nz to normalize the FFTs
			gridk_value.w = 6.0*3.1415926536 * (1.0 + k2/4.0/xisq) * expf( -(1-eta) * k2/4.0/xisq ) / ( k2 ) / Scalar( Nx*Ny*Nz );
		}

                gridk[tid] = gridk_value;

        }
}
