// Maintainer: Gang Wang
// Matrix wrappers written specific for CUSP library
// Adapted from codes of Andrew Fiore.

#include "hoomd/hoomd_config.h"
#include "Mobility.cuh"
#include "Helper.cuh"
#include <cusp/linear_operator.h>

#ifndef __MATRIX_WRAPPER_CUH__
#define __MATRIX_WRAPPER_CUH__

/*!
  cusp shell representing mobility tensor
  */
class cuspMobility : public cusp::linear_operator<Scalar, cusp::device_memory>{

public:
  // Pointer members
  Scalar4 *d_pos; //!< Position of particles
  unsigned int *d_group_members; //!< Index lookup table for group members
  Scalar4 *d_ewaldC1; //!< Ewald lookup table
  Scalar4 *d_gridk; //!< k for wavespace calculation
  CUFFTCOMPLEX *d_gridX; //!< ForceX in wavespace
  CUFFTCOMPLEX *d_gridY; //!< ForceY in wavespace
  CUFFTCOMPLEX *d_gridZ; //!< ForceZ in wavespace
  const unsigned int *d_n_neigh; //!< Number of neighbors
  const unsigned int *d_nlist; //!< Neighbor list
  const unsigned int *d_headlist;
  const Scalar *d_diameter; //!< Particle diameter

  // Variable members
  const unsigned int group_size; //!< Group size
  const BoxDim box; //!< Box, including some fast functions
  Scalar xi;
  Scalar eta;
  Scalar self;
  const int P;
  Scalar3 gridh;
  const Scalar ewald_cut; //!< Ewald cutoff
  const Scalar ewald_dr; //!< Ewald lookup table resolution
  const int ewald_n; //!< Ewald lookup table length
  cufftHandle plan; //!< CUFFT plan
  const int Nx; //!< NX in FFT
  const int Ny; //!< NY in FFT
  const int Nz; //!< NZ in FFT
  const unsigned int NxNyNz; //!< NX*NY*NZ

  const unsigned int N_total; //!< Total particle number (length of vector d_vel and d_net_force)

  const dim3 grid; //!< grid for kernel
  const dim3 threads; //!< threads for kernel
  const int gridBlockSize; //!< block size for wavespace calculation
  const int gridNBlock; //!< block grid for wavespace calculation

	// constructor
  cuspMobility(Scalar4 *d_pos,
    unsigned int *d_group_members,
    Scalar4 *d_ewaldC1,
    Scalar4 *d_gridk,
    CUFFTCOMPLEX *d_gridX,
    CUFFTCOMPLEX *d_gridY,
    CUFFTCOMPLEX *d_gridZ,
    const unsigned int *d_n_neigh,
    const unsigned int *d_nlist,
    const unsigned int *d_headlist,
    const Scalar *d_diameter,
    const unsigned int group_size,
    const BoxDim box,
    Scalar xi,
    Scalar eta,
    Scalar self,
    const int P,
    Scalar3 gridh,
    const Scalar ewald_cut,
    const Scalar ewald_dr,
    const int ewald_n,
    cufftHandle plan,
    const int Nx,
    const int Ny,
    const int Nz,
    const unsigned int NxNyNz,
    const unsigned int N_total,
    const dim3 grid,
    const dim3 threads,
    const int gridBlockSize,
    const int gridNBlock)
			:cusp::linear_operator<Scalar,cusp::device_memory>(4*N_total,4*N_total),
			d_pos(d_pos),
      d_group_members(d_group_members),
      d_ewaldC1(d_ewaldC1),
      d_gridk(d_gridk),
      d_gridX(d_gridX),
      d_gridY(d_gridY),
      d_gridZ(d_gridZ),
      d_n_neigh(d_n_neigh),
      d_nlist(d_nlist),
      d_headlist(d_headlist),
      d_diameter(d_diameter),
      group_size(group_size),
      box(box),
      xi(xi),
      eta(eta),
      self(self),
      P(P),
      gridh(gridh),
      ewald_cut(ewald_cut),
      ewald_dr(ewald_dr),
      ewald_n(ewald_n),
      plan(plan),
      Nx(Nx),
      Ny(Ny),
      Nz(Nz),
      NxNyNz(NxNyNz),
      N_total(N_total),
      grid(grid),
      threads(threads),
      gridBlockSize(gridBlockSize),
      gridNBlock(gridNBlock){}

	template<typename VectorType1, typename VectorType2>
		void operator()(VectorType1& x, VectorType2& y){
		Scalar *x_ptr = thrust::raw_pointer_cast(&x[0]);
		Scalar *y_ptr = thrust::raw_pointer_cast(&y[0]);

		Scalar4 *d_vel = (Scalar4 *) y_ptr;
		Scalar4 *d_net_force = (Scalar4 *) x_ptr;
    gpu_stokes_Mobility_wrap(d_pos,
      d_vel,
      d_net_force,
      d_group_members,
      group_size,
      box,
      xi,
      eta,
      ewald_cut,
      ewald_dr,
      ewald_n,
      d_ewaldC1,
      self,
      d_gridk,
      d_gridX,
      d_gridY,
      d_gridZ,
      plan,
      Nx,
      Ny,
      Nz,
      d_n_neigh,
      d_nlist,
      d_headlist,
      NxNyNz,
      grid,
      threads,
      gridBlockSize,
      gridNBlock,
      P,
      gridh,
      d_diameter );
	}
};

/*!
  cusp shell representing the "saddle" problem of rigid body hydrodynamic calculation
  */
class cuspRigidSaddle : public cusp::linear_operator<Scalar, cusp::device_memory>{
public:

  // Pointer members
  Scalar4 *d_vel;
  Scalar4 *d_pos;
  unsigned int *d_group_members;
  Scalar4 *d_ewaldC1;
  Scalar4 *d_gridk;
  CUFFTCOMPLEX *d_gridX;
  CUFFTCOMPLEX *d_gridY;
  CUFFTCOMPLEX *d_gridZ;
  const unsigned int *d_n_neigh;
  const unsigned int *d_nlist;
  const unsigned int *d_headlist;
  const Scalar *d_diameter;
  unsigned int *d_rigidindex;
  Scalar4 *d_rigidorientation;
  unsigned int *d_rigidparticleindices;
  Scalar4 *d_rigidparticlepos;

  // Variable members
  const unsigned int group_size;
  const BoxDim box;
  Scalar xi;
  Scalar eta;
  Scalar self;
  const int P;
  Scalar3 gridh;
  const Scalar ewald_cut;
  const Scalar ewald_dr;
  const int ewald_n;
  cufftHandle plan;
  const int Nx;
  const int Ny;
  const int Nz;
  const unsigned int NxNyNz;

  const unsigned int N_total;
  unsigned int n_bodies;
  unsigned int nmax;
  const unsigned int N_rigid_total;

  const dim3 grid;
  const dim3 threads;
  const int gridBlockSize;
  const int gridNBlock;
  const dim3 grid_rigid_particle;
  const dim3 threads_rigid_particle;
  const unsigned int rigidThreadDx;
  const unsigned int rigidThreadDy;

	cuspRigidSaddle(Scalar4 *d_vel,
    Scalar4 *d_pos,
    unsigned int *d_group_members,
    Scalar4 *d_ewaldC1,
    Scalar4 *d_gridk,
    CUFFTCOMPLEX *d_gridX,
    CUFFTCOMPLEX *d_gridY,
    CUFFTCOMPLEX *d_gridZ,
    const unsigned int *d_n_neigh,
    const unsigned int *d_nlist,
    const unsigned int *d_headlist,
    const Scalar *d_diameter,
    unsigned int *d_rigidindex,
    Scalar4 *d_rigidorientation,
    unsigned int *d_rigidparticleindices,
    Scalar4 *d_rigidparticlepos,
    const unsigned int group_size,
    const BoxDim box,
    Scalar xi,
    Scalar eta,
    Scalar self,
    const int P,
    Scalar3 gridh,
    const Scalar ewald_cut,
    const Scalar ewald_dr,
    const int ewald_n,
    cufftHandle plan,
    const int Nx,
    const int Ny,
    const int Nz,
    const unsigned int NxNyNz,
    const unsigned int N_total,
    unsigned int n_bodies,
    unsigned int nmax,
    const unsigned int N_rigid_total,
    const dim3 grid,
    const dim3 threads,
    const int gridBlockSize,
    const int gridNBlock,
    const dim3 grid_rigid_particle,
    const dim3 threads_rigid_particle,
    const unsigned int rigidThreadDx,
    const unsigned int rigidThreadDy)
			:cusp::linear_operator<Scalar,cusp::device_memory>(4*N_total+8*N_rigid_total,4*N_total+8*N_rigid_total),
      d_vel(d_vel),
      d_pos(d_pos),
      d_group_members(d_group_members),
      d_ewaldC1(d_ewaldC1),
      d_gridk(d_gridk),
      d_gridX(d_gridX),
      d_gridY(d_gridY),
      d_gridZ(d_gridZ),
      d_n_neigh(d_n_neigh),
      d_nlist(d_nlist),
      d_headlist(d_headlist),
      d_diameter(d_diameter),
      d_rigidindex(d_rigidindex),
      d_rigidorientation(d_rigidorientation),
      d_rigidparticleindices(d_rigidparticleindices),
      d_rigidparticlepos(d_rigidparticlepos),
      group_size(group_size),
      box(box),
      xi(xi),
      eta(eta),
      self(self),
      P(P),
      gridh(gridh),
      ewald_cut(ewald_cut),
      ewald_dr(ewald_dr),
      ewald_n(ewald_n),
      plan(plan),
      Nx(Nx),
      Ny(Ny),
      Nz(Nz),
      NxNyNz(NxNyNz),
      N_total(N_total),
      n_bodies(n_bodies),
      nmax(nmax),
      N_rigid_total(N_rigid_total),
      grid(grid),
      threads(threads),
      gridBlockSize(gridBlockSize),
      gridNBlock(gridNBlock),
      grid_rigid_particle(grid_rigid_particle),
      threads_rigid_particle(threads_rigid_particle),
      rigidThreadDx(rigidThreadDx),
      rigidThreadDy(rigidThreadDy){}

	template<typename VectorType1, typename VectorType2>
		void operator()(VectorType1& x, VectorType2& y){

		Scalar *x_ptr = thrust::raw_pointer_cast(&x[0]);
		Scalar *y_ptr = thrust::raw_pointer_cast(&y[0]);

    gpu_stokesrigid_RigidSaddle_wrap(x_ptr,
      y_ptr,
      d_vel,
      d_pos,
      d_group_members,
      d_ewaldC1,
      d_gridk,
      d_gridX,
      d_gridY,
      d_gridZ,
      d_n_neigh,
      d_nlist,
      d_headlist,
      d_diameter,
      d_rigidindex,
      d_rigidorientation,
      d_rigidparticleindices,
      d_rigidparticlepos,
      group_size,
      box,
      xi,
      eta,
      ewald_cut,
      ewald_dr,
      ewald_n,
      self,
      plan,
      Nx,
      Ny,
      Nz,
      NxNyNz,
      gridh,
      N_total,
      n_bodies,
      nmax,
      N_rigid_total,
      grid,
      threads,
      gridBlockSize,
      gridNBlock,
      P,
      grid_rigid_particle,
      threads_rigid_particle,
      rigidThreadDx,
      rigidThreadDy);

		}
};
#endif
