// Maintainer: Gang Wang
// Preconditioner for GMRES
#include "PreCond.cuh"

/*!
	Preconditioner helper kernel function

	d_z := (Preconditioner)^-1 * d_r
	Here, preconditioner is Sigma*Sigma^(T). (Mass and moment of inertia)
*/
__global__
void gpu_stokesrigid_InertiaInv_kernel(Scalar4 *d_r, Scalar4 *d_r_ang, Scalar4 *d_z, Scalar4 *d_z_ang, Scalar *d_rigidbodymass, Scalar4 *d_rigidmomentinertia, Scalar4 *d_rigidorientation, unsigned int group_size, unsigned int *d_group_members){
	int group_idx = blockDim.x * blockIdx.x + threadIdx.x;
	if (group_idx < group_size) {
		unsigned int idx = d_group_members[group_idx];

		Scalar4 r = d_r[idx];
		Scalar3 z = make_scalar3(r.x, r.y, r.z) / d_rigidbodymass[idx];
		d_z[idx] = make_scalar4(z.x, z.y, z.z, 0.0);

		Scalar4 r_ang = d_r_ang[idx];
		Scalar4 quat = d_rigidorientation[idx];
		Scalar4 momentinertia = d_rigidmomentinertia[idx];

		Scalar4 ex_space, ey_space, ez_space;
		exyzFromQuaternion(quat, ex_space, ey_space, ez_space);

		Scalar4 z_ang = make_scalar4(0.0, 0.0, 0.0, 0.0);

		computeAngularVelocity(r_ang, momentinertia, ex_space, ey_space, ez_space, z_ang);

		d_z_ang[idx] = z_ang;

	}
}

/*!
	Wrap all the functions to compute the saddle problem of rigid body motion
	Drive GPU kernel functions
*/
void gpu_stokesrigid_PreconditionerSigma_wrap(Scalar *x_ptr,
		Scalar *y_ptr,
		Scalar4 *d_temp,
		unsigned int *d_group_members,
		unsigned int *d_rigidindex,
		Scalar4 *d_rigidorientation,
		unsigned int *d_rigidparticleindices,
		Scalar4 *d_rigidparticlepos,
		Scalar *d_rigidbodymass,
		Scalar4 *d_rigidmomentinertia,
		const unsigned int group_size,
		const unsigned int N_total,
		unsigned int n_bodies,
		unsigned int nmax,
		const unsigned int N_rigid_total,
		const dim3 grid,
		const dim3 threads,
		const dim3 grid_rigid_particle,
    		const dim3 threads_rigid_particle,
		const unsigned int rigidThreadDx,
		const unsigned int rigidThreadDy,
		unsigned int grid_rigid,
		unsigned int thread_rigid,
		Scalar radius_rep){

	Scalar4 *x_P = (Scalar4 *) &x_ptr[0];
	Scalar4 *x_R1 = (Scalar4 *) &x_ptr[4*N_total];
	Scalar4 *x_R2 = (Scalar4 *) &x_ptr[4*N_total+4*N_rigid_total];

	Scalar4 *y_P = (Scalar4 *) &y_ptr[0];
	Scalar4 *y_R1 = (Scalar4 *) &y_ptr[4*N_total];
	Scalar4 *y_R2 = (Scalar4 *) &y_ptr[4*N_total+4*N_rigid_total];

	Scalar4 *d_tempP = &d_temp[0];
	Scalar4 *d_tempR1 = &d_temp[N_total];
	Scalar4 *d_tempR2 = &d_temp[N_total + N_rigid_total];

	// InertiaInv * x_R -> d_tempR
	gpu_stokesrigid_InertiaInv_kernel<<<grid_rigid, thread_rigid>>>(x_R1, x_R2, d_tempR1, d_tempR2, d_rigidbodymass, d_rigidmomentinertia, d_rigidorientation, n_bodies, d_rigidindex);

	// Sigma * x_P -> y_R
	gpu_stokesrigid_rigidsigma_kernel<<<grid_rigid_particle,threads_rigid_particle,rigidThreadDx*rigidThreadDy*sizeof(Scalar3)*2>>>(x_P,
				y_R1,
				y_R2,
				d_rigidindex,
				d_rigidorientation,
				n_bodies,
				nmax,
				d_rigidparticleindices,
				d_rigidparticlepos);

	// InertiaInv * y_R -> y_R
	gpu_stokesrigid_InertiaInv_kernel<<<grid_rigid, thread_rigid>>>(y_R1, y_R2, y_R1, y_R2, d_rigidbodymass, d_rigidmomentinertia, d_rigidorientation, n_bodies, d_rigidindex);

	// y_R - d_tempR -> y_R
	gpu_stokes_LinearCombination_kernel<<<grid_rigid, thread_rigid>>>(y_R1, d_tempR1, y_R1, 1.0, -1.0/radius_rep, n_bodies, d_rigidindex);
	gpu_stokes_LinearCombination_kernel<<<grid_rigid, thread_rigid>>>(y_R2, d_tempR2, y_R2, 1.0, -1.0/radius_rep, n_bodies, d_rigidindex);

	// Sigma^T * y_R -> d_tempP
	gpu_stokesrigid_rigidsigmatrans_kernel<<<grid_rigid_particle,threads_rigid_particle>>>(d_tempP,
				y_R1,
				y_R2,
				d_rigidindex,
				d_rigidorientation,
				n_bodies,
				nmax,
				d_rigidparticleindices,
				d_rigidparticlepos);

	// x_P - d_tempP -> y_P
	gpu_stokes_LinearCombination_kernel<<<grid, threads>>>(x_P, d_tempP, y_P, radius_rep, -radius_rep, group_size, d_group_members);

}
