// Modified by Gang Wang
// Modified based on new PSE scheme of Andrew Fiore

#ifdef WIN32
#pragma warning( push )
#pragma warning( disable : 4244 )
#endif

#include <boost/python.hpp>
using namespace boost::python;
#include <boost/bind.hpp>
using namespace boost;
#include <vector>
#include <algorithm>

#include "StokesRigid.h"
#include "StokesRigid.cuh"
#include <stdio.h>

/*! \file StokesRigid.cc
    \brief Contains code for the StokesRigid class
*/
/*! \param sysdef SystemDefinition this method will act on. Must not be NULL.
    \param group The group of particles this integration method is to work on
    \param T temperature Variant object
    \param seed the seed for RNG
    \param nlist shared_ptr to the NeighborList
    \param xi xi parameter for PSE code
    \param error error parameter for PSE code
    \param hydrostress boolean indicating if the stress is computed
    \param fname file name of the output stress file
*/
StokesRigid::StokesRigid(boost::shared_ptr<SystemDefinition> sysdef,
    boost::shared_ptr<ParticleGroup> group,
    boost::shared_ptr<Variant> T,
    unsigned int seed,
    boost::shared_ptr<NeighborList> nlist,
    Scalar xi,
    Scalar error,
    bool hydrostress,
    std::string fname
) : IntegrationMethodTwoStep(sysdef, group),
    m_T(T),
    m_seed(seed),
    m_nlist(nlist),
    m_xi(xi),
    m_error(error),
    m_hydrostress(hydrostress),
    m_hydrostress_period(100),
    m_delimiter("\t")
    {
        m_exec_conf->msg->notice(5) << "Constructing StokesRigid" << std::endl;

        // Hash the User's Seed to make it less likely to be a low positive integer
        m_seed = m_seed * 0x12345677 + 0x12345; m_seed ^= (m_seed >> 16); m_seed *= 0x45679;

        // only one GPU is supported
        if (!m_exec_conf->isCUDAEnabled())
        {
            m_exec_conf->msg->error() << "Creating a StokesRigid when CUDA is disabled" << std::endl;
            throw std::runtime_error("Error initializing StokesRigid");
        }

        // Get the system rigid data
        m_rigid_data = sysdef->getRigidData();
        // Create the rigid body group from the particle group
        m_body_group = boost::shared_ptr<RigidBodyGroup>(new RigidBodyGroup(sysdef, m_group));

        // Get the number of rigid bodies for frequent use
        m_n_bodies = m_body_group->getNumMembers();

        if (m_n_bodies == 0)
        {
            m_exec_conf->msg->warning() << "StokesRigid: Empty group." << std::endl;
        }

        m_hydrostress_now = false; // If m_hydrostress is not true, this will remain false.

        if (m_hydrostress){

            // open the file
            m_exec_conf->msg->notice(2) << "StokesRigid: Creating new hydrodynamic stress in file \"" << fname << "\"" << std::endl;
            m_file.open(fname.c_str(), std::ios_base::out);

            if (!m_file.good())
            {
                m_exec_conf->msg->error() << "StokesRigid: Unable to open file " << fname << std::endl;
                throw std::runtime_error("Error initializing StokesRigid");
            }

            writeHeader();

            m_hydrostress_now = true; // Compute stress at the first timestep.
        }

    }

StokesRigid::~StokesRigid()
    {
      m_exec_conf->msg->notice(5) << "Destroying StokesRigid" << std::endl;
      cufftDestroy(plan);
    }


/*!
	Set the parameters for Ewald summation
	And for Brownian force calculation
*/
void StokesRigid::setParams()
{

  // Try two Lanczos iterations to start (number of iterations will adapt as needed)
  m_m_Lanczos = 2;

  // Real space cutoff
  m_ewald_cut = sqrtf( - logf( m_error ) ) / m_xi;

  // Number of grid points
  int kmax = int( 2.0 * sqrtf( - logf( m_error ) ) * m_xi ) + 1;

  const BoxDim& box = m_pdata->getBox(); // Only for box not changing with time.
	Scalar3 L = box.getL();

  m_Nx = int( kmax * L.x / (2.0 * 3.1415926536 ) * 2.0 ) + 1;
  m_Ny = int( kmax * L.y / (2.0 * 3.1415926536 ) * 2.0 ) + 1;
  m_Nz = int( kmax * L.z / (2.0 * 3.1415926536 ) * 2.0 ) + 1;

  // Get list of int values between 8 and 512 that can be written as
  // 	(2^a)*(3^b)*(5^c)
  // Then sort list from low to high
  std::vector<int> Mlist;

  for ( int ii = 0; ii < 10; ++ii ){
    int pow2 = 1;
    for ( int i = 0; i < ii; ++i ){
      pow2 *= 2;
    }
    for ( int jj = 0; jj < 6; ++jj ){
      int pow3 = 1;
      for ( int j = 0; j < jj; ++j ){
        pow3 *= 3;
      }
      for ( int kk = 0; kk < 4; ++kk ){
        int pow5 = 1;
        for ( int k = 0; k < kk; ++k ){
          pow5 *= 5;
        }
        int Mcurr = pow2 * pow3 * pow5;
        if ( Mcurr >= 8 && Mcurr <= 512 ){
          Mlist.push_back(Mcurr);
        }
      }
    }
  }

  std::sort(Mlist.begin(), Mlist.end());
  const int nmult = Mlist.size();  // 62 such values exist

  // Compute the number of grid points in each direction
  //
  // Number of grid points should be a power of 2,3,5 for most efficient FFTs
  for ( int ii = 0; ii < nmult; ++ii ){
    if (m_Nx <= Mlist[ii]){
       m_Nx = Mlist[ii];
      break;
    }
  }
  for ( int ii = 0; ii < nmult; ++ii ){
    if (m_Ny <= Mlist[ii]){
      m_Ny = Mlist[ii];
      break;
    }
  }
  for ( int ii = 0; ii < nmult; ++ii ){
    if (m_Nz <= Mlist[ii]){
      m_Nz = Mlist[ii];
      break;
    }
  }

  if ( m_Nx > 512 || m_Ny > 512 || m_Nz > 512 ){

    printf("Requested Number of Fourier Nodes Exceeds Max Dimension of 512\n");
    printf("Mx = %i \n", m_Nx);
    printf("My = %i \n", m_Ny);
    printf("Mz = %i \n", m_Nz);

    exit(EXIT_FAILURE);
  }

  // Maximum eigenvalue of A'*A to scale P
  Scalar gamma = m_max_strain; // What is this value?
  Scalar gamma2 = gamma*gamma;
  Scalar lambda = 1.0 + gamma2/2.0 + gamma*sqrtf(1.0 + gamma2/4.0);

  // Grid spacing
  m_gridh = L / make_scalar3(m_Nx,m_Ny,m_Nz);

  // Parameters for the Spectral Ewald Method (Lindbo and Tornberg, J. Comp. Phys., 2011)
  m_gaussm = 1.0;
  while ( erfcf( m_gaussm / sqrtf(2.0*lambda) ) > m_error ){
      m_gaussm = m_gaussm + 0.01;
  }
  m_gaussP = int( m_gaussm*m_gaussm / 3.1415926536 )  + 1;

  if (m_gaussP > m_Nx) m_gaussP = m_Nx; // Can't be supported beyond grid
  if (m_gaussP > m_Ny) m_gaussP = m_Ny;
  if (m_gaussP > m_Nz) m_gaussP = m_Nz;
  Scalar w = m_gaussP*m_gridh.x / 2.0;	               // Gaussian width in simulation units
  Scalar xisq  = m_xi * m_xi;
  m_eta = (2.0*w/m_gaussm)*(2.0*w/m_gaussm) * ( xisq );  // Gaussian splitting parameter

  // Print summary to command line output
  printf("\n");
  printf("\n");
  m_exec_conf->msg->notice(2) << "=== StokesRigid Version 0.5, Jun. 22 2017 (SAME RADIUS) ===" << std::endl; // Radius is always the same?
  m_exec_conf->msg->notice(2) << "--- NUFFT Hydrodynamics Statistics ---" << std::endl;
  m_exec_conf->msg->notice(2) << "Mx: " << m_Nx << std::endl;
  m_exec_conf->msg->notice(2) << "My: " << m_Ny << std::endl;
  m_exec_conf->msg->notice(2) << "Mz: " << m_Nz << std::endl;
  m_exec_conf->msg->notice(2) << "rcut: " << m_ewald_cut << std::endl;
  m_exec_conf->msg->notice(2) << "Points per radius (x,y,z): " << m_Nx / L.x << ", " << m_Ny / L.y << ", " << m_Nz / L.z << std::endl;
  m_exec_conf->msg->notice(2) << "--- Gaussian Spreading Parameters ---"  << std::endl;
  m_exec_conf->msg->notice(2) << "gauss_m: " << m_gaussm << std::endl;
  m_exec_conf->msg->notice(2) << "gauss_P: " << m_gaussP << std::endl;
  m_exec_conf->msg->notice(2) << "gauss_eta: " << m_eta << std::endl;
  m_exec_conf->msg->notice(2) << "gauss_w: " << w << std::endl;
  m_exec_conf->msg->notice(2) << "gauss_gridh (x,y,z): " << L.x/m_Nx << ", " << L.y/m_Ny << ", " << L.z/m_Nz << std::endl;
  printf("\n");
  printf("\n");

	cufftPlan3d(&plan, m_Nx, m_Ny, m_Nz, CUFFT_C2C);

	GPUArray<Scalar4> n_gridk(m_Nx*m_Ny*m_Nz, m_exec_conf);
	m_gridk.swap(n_gridk);
	GPUArray<CUFFTCOMPLEX> n_gridX(m_Nx*m_Ny*m_Nz, m_exec_conf);
	m_gridX.swap(n_gridX);
	GPUArray<CUFFTCOMPLEX> n_gridY(m_Nx*m_Ny*m_Nz, m_exec_conf);
	m_gridY.swap(n_gridY);
	GPUArray<CUFFTCOMPLEX> n_gridZ(m_Nx*m_Ny*m_Nz, m_exec_conf);
	m_gridZ.swap(n_gridZ);

  // Get list of reciprocal space vectors, and scaling factor for the wave space calculation at each grid point
  ArrayHandle<Scalar4> h_gridk(m_gridk, access_location::host, access_mode::readwrite);
  for (int i = 0; i < m_Nx; i++) {
    for (int j = 0; j < m_Ny; j++) {
      for (int k = 0; k < m_Nz; k++) {

        // Index into grid vector storage array
        int idx = i * m_Ny*m_Nz + j * m_Nz + k;

        // k goes from -N/2 to N/2
        h_gridk.data[idx].x = 2.0*3.1415926536 * ((i < ( m_Nx + 1 ) / 2) ? i : i - m_Nx) / L.x;
        h_gridk.data[idx].y = 2.0*3.1415926536 * ((j < ( m_Ny + 1 ) / 2) ? j : j - m_Ny) / L.y;
        h_gridk.data[idx].z = 2.0*3.1415926536 * ((k < ( m_Nz + 1 ) / 2) ? k : k - m_Nz) / L.z;

        // k dot k
        Scalar k2 = h_gridk.data[idx].x*h_gridk.data[idx].x + h_gridk.data[idx].y*h_gridk.data[idx].y + h_gridk.data[idx].z*h_gridk.data[idx].z;

        // Scaling factor used in wave space sum
        //
        // Can't include k=0 term in the Ewald sum
        if (i == 0 && j == 0 && k == 0){
          h_gridk.data[idx].w = 0;
        }
        else{
          // Have to divide by Nx*Ny*Nz to normalize the FFTs
          h_gridk.data[idx].w = 6.0*3.1415926536 * (1.0 + k2/4.0/xisq) * expf( -(1-m_eta) * k2/4.0/xisq ) / ( k2 ) / Scalar( m_Nx*m_Ny*m_Nz );
        }

      }
    }
  }

  // Store the coefficients for the real space part of Ewald summation
  //
  // Will precompute scaling factors for real space component of summation for a given
  //     discretization to speed up GPU calculations
  m_ewald_dr = 0.001; 		           // Distance resolution
  m_ewald_n = m_ewald_cut / m_ewald_dr - 1;  // Number of entries in tabulation

  double dr = 0.0010000000000000;

  // Get particle diameter and self bit
  ArrayHandle<Scalar> h_diameter(m_pdata->getDiameters(), access_location::host, access_mode::read);
  Scalar diameter = h_diameter.data[0];

  Scalar pi12 = 1.77245385091;
  Scalar aa = diameter / 2.0;
  Scalar axi = aa * m_xi;
  Scalar axi2 = axi * axi;
  m_self = (1. + 4.*pi12*axi*erfc(2.*axi) - exp(-4.*axi2))/(4.*pi12*axi*aa);

  // Allocate storage for real space Ewald table
  int nR = m_ewald_n + 1; // number of entries in ewald table
  GPUArray<Scalar4> n_ewaldC1( nR, m_exec_conf);
  m_ewaldC1.swap(n_ewaldC1);
  ArrayHandle<Scalar4> h_ewaldC1(m_ewaldC1, access_location::host, access_mode::readwrite);

  // Functions are complicated so calculation should be done in double precision, then truncated to single precision
  // in order to ensure accurate evaluation
  double xi  = m_xi;
  double Pi = 3.141592653589793;
  double a = aa;

  // Fill tables
  for ( int kk = 0; kk < nR; kk++ )
  {

    // Initialize entries
    h_ewaldC1.data[ kk ].x = 0.0; // UF1 at r
    h_ewaldC1.data[ kk ].y = 0.0; // UF2 at r
    h_ewaldC1.data[ kk ].z = 0.0; // UF1 at r + dr
    h_ewaldC1.data[ kk ].w = 0.0; // UF2 at r + dr

    // Distance for current entry
    double r = double( kk ) * dr + dr;
    double Imrr = 0, rr = 0;

    // Expression have been simplified assuming no overlap, touching, and overlap
    if ( r > 2.0*a ){

      Imrr = -pow(a,-1) + (pow(a,2)*pow(r,-3))/2. + (3*pow(r,-1))/4. + (3*erfc(r*xi)*pow(a,-2)*pow(r,-3)*(-12*pow(r,4) + pow(xi,-4)))/128. +
      pow(a,-2)*((9*r)/32. - (3*pow(r,-3)*pow(xi,-4))/128.) +
      (erfc((2*a + r)*xi)*(128*pow(a,-1) + 64*pow(a,2)*pow(r,-3) + 96*pow(r,-1) + pow(a,-2)*(36*r - 3*pow(r,-3)*pow(xi,-4))))/256. +
      (erfc(2*a*xi - r*xi)*(128*pow(a,-1) - 64*pow(a,2)*pow(r,-3) - 96*pow(r,-1) + pow(a,-2)*(-36*r + 3*pow(r,-3)*pow(xi,-4))))/
      256. + (3*exp(-(pow(r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-2)*pow(xi,-3)*(1 + 6*pow(r,2)*pow(xi,2)))/64. +
      (exp(-(pow(2*a + r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-3)*pow(xi,-3)*
      (8*r*pow(a,2)*pow(xi,2) - 16*pow(a,3)*pow(xi,2) + a*(2 - 28*pow(r,2)*pow(xi,2)) - 3*(r + 6*pow(r,3)*pow(xi,2))))/128. +
      (exp(-(pow(-2*a + r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-3)*pow(xi,-3)*
      (8*r*pow(a,2)*pow(xi,2) + 16*pow(a,3)*pow(xi,2) + a*(-2 + 28*pow(r,2)*pow(xi,2)) - 3*(r + 6*pow(r,3)*pow(xi,2))))/128.;

      rr = -pow(a,-1) - pow(a,2)*pow(r,-3) + (3*pow(r,-1))/2. + (3*pow(a,-2)*pow(r,-3)*(4*pow(r,4) + pow(xi,-4)))/64. +
      (erfc(2*a*xi - r*xi)*(64*pow(a,-1) + 64*pow(a,2)*pow(r,-3) - 96*pow(r,-1) + pow(a,-2)*(-12*r - 3*pow(r,-3)*pow(xi,-4))))/128. +
      (erfc((2*a + r)*xi)*(64*pow(a,-1) - 64*pow(a,2)*pow(r,-3) + 96*pow(r,-1) + pow(a,-2)*(12*r + 3*pow(r,-3)*pow(xi,-4))))/128. +
      (3*exp(-(pow(r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-2)*pow(xi,-3)*(-1 + 2*pow(r,2)*pow(xi,2)))/32. -
      ((2*a + 3*r)*exp(-(pow(-2*a + r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-3)*pow(xi,-3)*
      (-1 - 8*a*r*pow(xi,2) + 8*pow(a,2)*pow(xi,2) + 2*pow(r,2)*pow(xi,2)))/64. +
      ((2*a - 3*r)*exp(-(pow(2*a + r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-3)*pow(xi,-3)*
      (-1 + 8*a*r*pow(xi,2) + 8*pow(a,2)*pow(xi,2) + 2*pow(r,2)*pow(xi,2)))/64. -
      (3*erfc(r*xi)*pow(a,-2)*pow(r,-3)*pow(xi,-4)*(1 + 4*pow(r,4)*pow(xi,4)))/64.;

    }
    else if ( r == 2.0*a ){

      Imrr = -(pow(a,-5)*(3 + 16*a*xi*pow(Pi,-0.5))*pow(xi,-4))/2048. + (3*erfc(2*a*xi)*pow(a,-5)*(-192*pow(a,4) + pow(xi,-4)))/1024. +
      erfc(4*a*xi)*(pow(a,-1) - (3*pow(a,-5)*pow(xi,-4))/2048.) +
      (exp(-16*pow(a,2)*pow(xi,2))*pow(a,-4)*pow(Pi,-0.5)*pow(xi,-3)*(-1 - 64*pow(a,2)*pow(xi,2)))/256. +
      (3*exp(-4*pow(a,2)*pow(xi,2))*pow(a,-4)*pow(Pi,-0.5)*pow(xi,-3)*(1 + 24*pow(a,2)*pow(xi,2)))/256.;

      rr = (pow(a,-5)*(3 + 16*a*xi*pow(Pi,-0.5))*pow(xi,-4))/1024. + erfc(2*a*xi)*((-3*pow(a,-1))/8. - (3*pow(a,-5)*pow(xi,-4))/512.) +
      erfc(4*a*xi)*(pow(a,-1) + (3*pow(a,-5)*pow(xi,-4))/1024.) +
      (exp(-16*pow(a,2)*pow(xi,2))*pow(a,-4)*pow(Pi,-0.5)*pow(xi,-3)*(1 - 32*pow(a,2)*pow(xi,2)))/128. +
      (3*exp(-4*pow(a,2)*pow(xi,2))*pow(a,-4)*pow(Pi,-0.5)*pow(xi,-3)*(-1 + 8*pow(a,2)*pow(xi,2)))/128.;

    }
    else if ( r < 2*a){

      Imrr = (-9*r*pow(a,-2))/32. + pow(a,-1) - (pow(a,2)*pow(r,-3))/2. - (3*pow(r,-1))/4. +
      (3*erfc(r*xi)*pow(a,-2)*pow(r,-3)*(-12*pow(r,4) + pow(xi,-4)))/128. +
      (erfc((-2*a + r)*xi)*(-128*pow(a,-1) + 64*pow(a,2)*pow(r,-3) + 96*pow(r,-1) + pow(a,-2)*(36*r - 3*pow(r,-3)*pow(xi,-4))))/
      256. + (erfc((2*a + r)*xi)*(128*pow(a,-1) + 64*pow(a,2)*pow(r,-3) + 96*pow(r,-1) + pow(a,-2)*(36*r - 3*pow(r,-3)*pow(xi,-4))))/
      256. + (3*exp(-(pow(r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-2)*pow(xi,-3)*(1 + 6*pow(r,2)*pow(xi,2)))/64. +
      (exp(-(pow(2*a + r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-3)*pow(xi,-3)*
      (8*r*pow(a,2)*pow(xi,2) - 16*pow(a,3)*pow(xi,2) + a*(2 - 28*pow(r,2)*pow(xi,2)) - 3*(r + 6*pow(r,3)*pow(xi,2))))/128. +
      (exp(-(pow(-2*a + r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-3)*pow(xi,-3)*
      (8*r*pow(a,2)*pow(xi,2) + 16*pow(a,3)*pow(xi,2) + a*(-2 + 28*pow(r,2)*pow(xi,2)) - 3*(r + 6*pow(r,3)*pow(xi,2))))/128.;

      rr = ((2*a + 3*r)*pow(a,-2)*pow(2*a - r,3)*pow(r,-3))/16. +
      (erfc((-2*a + r)*xi)*(-64*pow(a,-1) - 64*pow(a,2)*pow(r,-3) + 96*pow(r,-1) + pow(a,-2)*(12*r + 3*pow(r,-3)*pow(xi,-4))))/128. +
      (erfc((2*a + r)*xi)*(64*pow(a,-1) - 64*pow(a,2)*pow(r,-3) + 96*pow(r,-1) + pow(a,-2)*(12*r + 3*pow(r,-3)*pow(xi,-4))))/128. +
      (3*exp(-(pow(r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-2)*pow(xi,-3)*(-1 + 2*pow(r,2)*pow(xi,2)))/32. -
      ((2*a + 3*r)*exp(-(pow(-2*a + r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-3)*pow(xi,-3)*
      (-1 - 8*a*r*pow(xi,2) + 8*pow(a,2)*pow(xi,2) + 2*pow(r,2)*pow(xi,2)))/64. +
      ((2*a - 3*r)*exp(-(pow(2*a + r,2)*pow(xi,2)))*pow(a,-2)*pow(Pi,-0.5)*pow(r,-3)*pow(xi,-3)*
      (-1 + 8*a*r*pow(xi,2) + 8*pow(a,2)*pow(xi,2) + 2*pow(r,2)*pow(xi,2)))/64. -
      (3*erfc(r*xi)*pow(a,-2)*pow(r,-3)*pow(xi,-4)*(1 + 4*pow(r,4)*pow(xi,4)))/64.;

    }

    // Save values to table
    h_ewaldC1.data[ kk ].x = Scalar( Imrr ); // UF1
    h_ewaldC1.data[ kk ].y = Scalar( rr );   // UF2

  } // kk loop over distances

  // Both pieces of UF data for faster interpolation (r and r+dr stored in same Scalar4)
  for ( int kk = 0; kk < (nR-1); kk++ ){

    int offset1 = kk;
    int offset2 = (kk+1);

    h_ewaldC1.data[ offset1 ].z = h_ewaldC1.data[ offset2 ].x;
    h_ewaldC1.data[ offset1 ].w = h_ewaldC1.data[ offset2 ].y;

  }

	// For stress calculation
	if (m_hydrostress){

		unsigned int N_rigid_total = m_rigid_data->getNumBodies();

		GPUArray<Scalar3> n_vecSii(N_rigid_total, m_exec_conf);
		GPUArray<Scalar3> n_vecSij(N_rigid_total, m_exec_conf);

		m_vecSii.swap(n_vecSii);
		m_vecSij.swap(n_vecSij);
	}

}


/*!
Helper function to calculate grid points for FFT
*/
unsigned int StokesRigid::nextPowerOf2(unsigned int n)
{
	n--;
	n |= n >> 1;
	n |= n >> 2;
	n |= n >> 4;
	n |= n >> 8;
	n |= n >> 16;
	n++;
	return n;
}


/*! \param timestep Current time step
\post Particle positions and velocities are moved forward to timestep+1
*/
void StokesRigid::integrateStepOne(unsigned int timestep)
{
	m_nlist->compute(timestep);

	// access the neighbor list
	ArrayHandle<unsigned int> d_n_neigh(m_nlist->getNNeighArray(), access_location::device, access_mode::read);
	ArrayHandle<unsigned int> d_nlist(m_nlist->getNListArray(), access_location::device, access_mode::read);
	ArrayHandle<unsigned int> d_headlist(m_nlist->getHeadList(), access_location::device, access_mode::read);

	unsigned int group_size = m_group->getNumMembers();
	assert(group_size <= m_pdata->getN());
	if (group_size == 0)
		return;

	const GPUArray< Scalar4 >& net_force = m_pdata->getNetForce();

	// profile this step
	if (m_prof)
		m_prof->push(m_exec_conf, "StokesRigid step 1: mobility");

	// access all the needed data
	ArrayHandle<Scalar4> d_pos(m_pdata->getPositions(), access_location::device, access_mode::readwrite);
	ArrayHandle<Scalar4> d_vel(m_pdata->getVelocities(), access_location::device, access_mode::readwrite);
	ArrayHandle<Scalar4> d_net_force(net_force, access_location::device, access_mode::read);
	ArrayHandle<int3> d_image(m_pdata->getImages(), access_location::device, access_mode::readwrite);

	// Read diameters for calculation
	ArrayHandle<Scalar> d_diameter(m_pdata->getDiameters(), access_location::device, access_mode::read);

	BoxDim box = m_pdata->getBox();
	ArrayHandle< unsigned int > d_index_array(m_group->getIndexArray(), access_location::device, access_mode::read);

	ArrayHandle<Scalar4> d_gridk(m_gridk, access_location::device, access_mode::readwrite);
	ArrayHandle<CUFFTCOMPLEX> d_gridX(m_gridX, access_location::device, access_mode::readwrite);
	ArrayHandle<CUFFTCOMPLEX> d_gridY(m_gridY, access_location::device, access_mode::readwrite);
	ArrayHandle<CUFFTCOMPLEX> d_gridZ(m_gridZ, access_location::device, access_mode::readwrite);

	ArrayHandle<Scalar4> d_ewaldC1(m_ewaldC1, access_location::device, access_mode::read);

	// Access needed rigid data
	ArrayHandle<Scalar4> d_rigidvel(m_rigid_data->getVel(), access_location::device, access_mode::readwrite);
	ArrayHandle<Scalar4> d_rigidangvel(m_rigid_data->getAngVel(), access_location::device, access_mode::readwrite);
	//ArrayHandle<Scalar4> d_rigidforce(m_rigid_data->getForce(), access_location::device, access_mode::readwrite);;
	//ArrayHandle<Scalar4> d_rigidtorque(m_rigid_data->getTorque(), access_location::device, access_mode::readwrite);;
	ArrayHandle<Scalar4> d_rigidcom(m_rigid_data->getCOM(), access_location::device, access_mode::readwrite);
	ArrayHandle<Scalar4> d_rigidorientation(m_rigid_data->getOrientation(), access_location::device, access_mode::readwrite);
	ArrayHandle<int3> d_rigidimage(m_rigid_data->getBodyImage(), access_location::device, access_mode::readwrite);
	ArrayHandle<Scalar4> d_rigidparticlepos(m_rigid_data->getParticlePos(), access_location::device, access_mode::read);
	ArrayHandle<unsigned int> d_rigidparticleindices(m_rigid_data->getParticleIndices(), access_location::device, access_mode::read);

	ArrayHandle<unsigned int> d_rigidindex(m_body_group->getIndexArray(), access_location::device, access_mode::read);

	// For preconditioning
	ArrayHandle<Scalar> d_rigidbodymass(m_rigid_data->getBodyMass(), access_location::device, access_mode::read);
	ArrayHandle<Scalar4> d_rigidmomentinertia(m_rigid_data->getMomentInertia(), access_location::device, access_mode::read);
	// For preconditioning

	unsigned int nmax = m_rigid_data->getNmax(); // Maximum number of particles per rigid body
	unsigned int rigidThreadDx; // Thread block dimension of X
	if (nmax > 512){
		rigidThreadDx = 512;
	}
	else{
		rigidThreadDx = nextPowerOf2(nmax);
	}
	unsigned int rigidThreadDy = 512 / rigidThreadDx;
	unsigned int rigidGrid = m_n_bodies / rigidThreadDy + 1;
	// 1D grid of blocks, 2D block of threads

	// Stress values
	Scalar3 SBii_total = make_scalar3(0.0, 0.0, 0.0); // Total stress from Brownian, ii part, 11, 22, 33
	Scalar3 SBij_total = make_scalar3(0.0, 0.0, 0.0); // Total stress from Brownian, ij part, 12, 13, 23

	Scalar3 SPii_total = make_scalar3(0.0, 0.0, 0.0); // Total stress from Force and Hydrodynamics (shear), ii part, 11, 22, 33
	Scalar3 SPij_total = make_scalar3(0.0, 0.0, 0.0); // Total stress from Force and Hydrodynamics (shear), ij part, 12, 13, 23

	// For stress calculation
	// Hopefully this doesn't cause any trouble.. NULL should be returned if stress is NOT computed
	ArrayHandle<Scalar3> d_vecSii(m_vecSii, access_location::device, access_mode::readwrite);
	ArrayHandle<Scalar3> d_vecSij(m_vecSij, access_location::device, access_mode::readwrite);

	const GPUArray<Scalar>& net_virial = m_pdata->getNetVirial();
	ArrayHandle<Scalar> d_net_virial(net_virial, access_location::device, access_mode::read);
	unsigned int virial_pitch = net_virial.getPitch();
	// For stress calculation

    // Calculate the shear rate of the current timestep
	// Scalar current_shear_rate = m_shear_rate * cos(m_shear_freq * 2 * 3.1415926536 * Scalar(timestep - m_offset) * m_deltaT);
    Scalar current_shear_rate = m_shear_func -> getShearRate(timestep);

	// divM test start
	Scalar3 test[4];
	// divM test end

	// perform the update on the GPU
    gpu_stokesrigid_step_one(d_pos.data,
        d_vel.data,
        d_index_array.data,
        group_size,
        box,
        m_deltaT,
        d_net_force.data,
        m_T->getValue(timestep),
        timestep,
        m_seed,
        m_xi,
        m_eta,
        m_self,
        m_gaussP,
        m_gridh,
        m_ewald_cut,
        m_ewald_dr,
        m_ewald_n,
        m_error,
        d_ewaldC1.data,
        m_m_Lanczos,
        d_gridk.data,
        d_gridX.data,
        d_gridY.data,
        d_gridZ.data,
        plan,
        m_Nx,
        m_Ny,
        m_Nz,
        d_n_neigh.data,
        d_nlist.data,
        d_headlist.data,
        m_pdata->getN(),
        d_diameter.data,
        d_rigidvel.data,
        d_rigidangvel.data,
        //d_rigidforce.data,
        //d_rigidtorque.data,
        d_rigidindex.data,
        d_rigidorientation.data,
        m_n_bodies,
        nmax,
        d_rigidparticleindices.data,
        d_rigidparticlepos.data,
        d_rigidcom.data,
        d_rigidimage.data,
        rigidThreadDx,
        rigidThreadDy,
        rigidGrid,
        m_rigid_data->getNumBodies(),
        d_rigidbodymass.data,
        d_rigidmomentinertia.data,
        d_image.data,
        m_hydrostress_now,
        d_vecSii.data,
        d_vecSij.data,
        &SBii_total,
        &SBij_total,
        &SPii_total,
        &SPij_total,
        d_net_virial.data,
        virial_pitch,
        current_shear_rate,
        test);

	// divM test start
/*
	std::ofstream output("AvgVel.txt", std::ofstream::app);
	output << test[0].x << " " << test[0].y << " " << test[0].z << " " << test[1].x << " " << test[1].y << " " << test[1].z << " ";
	output << test[2].x << " " << test[2].y << " " << test[2].z << " " << test[3].x << " " << test[3].y << " " << test[3].z;
	output << std::endl;
	output.close();
*/
	// divM test end

	if (m_exec_conf->isCUDAErrorCheckingEnabled())
		CHECK_CUDA_ERROR();

	if (m_hydrostress_now){

		Scalar3 L = box.getL();
		Scalar box_volume = L.x * L.y * L.z;

		SBii_total = SBii_total / box_volume;
		SBij_total = SBij_total / box_volume;

		SPii_total = SPii_total / box_volume;
		SPij_total = SPij_total / box_volume;

		// Write to the file
		m_file << std::setprecision(10) << timestep << m_delimiter;

		m_file << std::setprecision(10) << SBii_total.x << m_delimiter;
		m_file << std::setprecision(10) << SBii_total.y << m_delimiter;
		m_file << std::setprecision(10) << SBii_total.z << m_delimiter;

		m_file << std::setprecision(10) << SBij_total.x << m_delimiter;
		m_file << std::setprecision(10) << SBij_total.y << m_delimiter;
		m_file << std::setprecision(10) << SBij_total.z << m_delimiter;

		m_file << std::setprecision(10) << SPii_total.x << m_delimiter;
		m_file << std::setprecision(10) << SPii_total.y << m_delimiter;
		m_file << std::setprecision(10) << SPii_total.z << m_delimiter;

		m_file << std::setprecision(10) << SPij_total.x << m_delimiter;
		m_file << std::setprecision(10) << SPij_total.y << m_delimiter;
		m_file << std::setprecision(10) << SPij_total.z << std::endl;

		m_file.flush();

		if (!m_file.good())
		{
			m_exec_conf->msg->error() << "StokesRigid: I/O error while writing file" << std::endl;
			throw std::runtime_error("Error writting file");
		}
	}

	// Set boolean variable for stress calculation correctly
	if (m_hydrostress){
		if (m_hydrostress_now){
			m_hydrostress_now = false;
			m_hydrostress_lastcompute = timestep;
		}
		if (m_hydrostress_lastcompute + m_hydrostress_period - 1 == timestep){
			m_hydrostress_now = true;
		}
	}

	// done profiling
	if (m_prof)
		m_prof->pop(m_exec_conf);

}

/*! \param timestep Current time step
\post Nothing is done.
*/
void StokesRigid::integrateStepTwo(unsigned int timestep)
{
}

/*!
Checks that every particle in the group is valid. This method may be called by anyone wishing to make this error check.
Check here that all particles belong to rigid bodies.
*/
void StokesRigid::validateGroup()
{
	for (unsigned int gidx = 0; gidx < m_group->getNumMembers(); gidx++)
	{
		unsigned int tag = m_group->getMemberTag(gidx);
		if (m_pdata->getBody(tag) == NO_BODY)
		{
			m_exec_conf->msg->error() << "StokesRigid: Particle " << tag << " does not belong to a rigid body. "
				<< "This integration method does not operate on free particles." << std::endl;

			throw std::runtime_error("Error initializing integration method");
		}
	}
}

/*! \param query_group Group over which to count degrees of freedom.
A majority of the integration methods add D degrees of freedom per particle in \a query_group that is also in the
group assigned to the method. Hence, StokesRigid will implement that counting.
*/
unsigned int StokesRigid::getNDOF(boost::shared_ptr<ParticleGroup> query_group)
{
	ArrayHandle<Scalar4> moment_inertia_handle(m_rigid_data->getMomentInertia(), access_location::host, access_mode::read);

	// count the number of particles both in query_group and m_group
	boost::shared_ptr<ParticleGroup> intersect_particles = ParticleGroup::groupIntersection(m_group, query_group);

	RigidBodyGroup intersect_bodies(m_sysdef, intersect_particles);

	// Counting body DOF:
	// 3D systems: a body has 6 DOF by default, subtracted by the number of zero moments of inertia
	// 2D systems: a body has 3 DOF by default
	unsigned int query_group_dof = 0;
	unsigned int dimension = m_sysdef->getNDimensions();
	unsigned int dof_one;
	for (unsigned int group_idx = 0; group_idx < intersect_bodies.getNumMembers(); group_idx++)
	{
		unsigned int body = intersect_bodies.getMemberIndex(group_idx);
		if (m_body_group->isMember(body))
		{
			if (dimension == 3)
			{
				dof_one = 6;
				if (moment_inertia_handle.data[body].x == 0.0)
					dof_one--;

				if (moment_inertia_handle.data[body].y == 0.0)
					dof_one--;

				if (moment_inertia_handle.data[body].z == 0.0)
					dof_one--;
			}
			else
			{
				dof_one = 3;
				if (moment_inertia_handle.data[body].z == 0.0)
					dof_one--;
			}

			query_group_dof += dof_one;
		}
	}

	return query_group_dof;
}


/*! The entire header row is written to the file. First, timestep is written as every file includes it and then the
columns are looped through and their names printed, separated by the delimiter.
*/
void StokesRigid::writeHeader()
{

	m_file << "timestep";

	m_file << m_delimiter << "SB_xx";
	m_file << m_delimiter << "SB_yy";
	m_file << m_delimiter << "SB_zz";
	m_file << m_delimiter << "SB_xy";
	m_file << m_delimiter << "SB_xz";
	m_file << m_delimiter << "SB_yz";

	m_file << m_delimiter << "SP_xx";
	m_file << m_delimiter << "SP_yy";
	m_file << m_delimiter << "SP_zz";
	m_file << m_delimiter << "SP_xy";
	m_file << m_delimiter << "SP_xz";
	m_file << m_delimiter << "SP_yz";

	m_file << std::endl;

	m_file.flush();
}


// This seems very important...
PDataFlags StokesRigid::getRequestedPDataFlags()
{

	PDataFlags flags(0);

	if (m_hydrostress){
		flags[pdata_flag::pressure_tensor] = 1;
	}

	return flags;

}

void export_StokesRigid()
    {
        class_<StokesRigid, boost::shared_ptr<StokesRigid>, bases<IntegrationMethodTwoStep>, boost::noncopyable>
        ("StokesRigid", init< boost::shared_ptr<SystemDefinition>, boost::shared_ptr<ParticleGroup>, boost::shared_ptr<Variant>, unsigned int, boost::shared_ptr<NeighborList>, Scalar, Scalar, bool, std::string >())
        .def("setT", &StokesRigid::setT)
        .def("setParams", &StokesRigid::setParams)
        .def("setDelimiter", &StokesRigid::setDelimiter)
        .def("setHydrostressPeriod", &StokesRigid::setHydrostressPeriod)
        .def("setShear", &StokesRigid::setShear)
        ;
    }

#ifdef WIN32
#pragma warning( pop )
#endif
