// Modified by Gang Wang
// Modified based on new PSE scheme of Andrew Fiore

#include "StokesRigid.cuh"
#include "saruprngCUDA.h"
#include <cusp/monitor.h>
#include <cusp/krylov/gmres.h>
#include <cusp/krylov/cg.h>
#include <cusp/array1d.h>

#define KRYLOV_ERROR 1e-6
#define RFD_DELTA 0.01
#define GMRES_RESET 200
#define GMRES_ITER_NUM 5000

//! Texture for reading table values
scalar4_tex_t tables1_tex;
//! Texture for reading particle positions
scalar4_tex_t pos_tex;

//! Shared memory array for rigid body calculation
extern __shared__ Scalar3 shared_rigid_vel[];

/*! \file StokesRigid.cu
    \brief Defines GPU kernel code for integration considering hydrodynamic interactions and rigid bodies on the GPU. Used by StokesRigid.cc.
*/

//! Kernel function to set d_pos of the particles
__global__
void gpu_stokesrigid_setdpos_kernel(Scalar4 *d_pos,
  int3 *d_image,
  unsigned int *d_rigidindex,
  Scalar4 *d_rigidorientation,
  unsigned int n_bodies,
  unsigned int nmax,
  unsigned int *d_rigidparticleindices,
  Scalar4 *d_rigidparticlepos,
  Scalar4 *d_rigidcom,
  int3 *d_rigidimage,
  const BoxDim box){

	unsigned int body_group_idx = blockIdx.x * blockDim.y + threadIdx.y;
	unsigned int group_idx = threadIdx.x;
	unsigned int body_idx;
	unsigned int idx;

	if (body_group_idx < n_bodies){
		body_idx = d_rigidindex[body_group_idx];

		Scalar4 quat = d_rigidorientation[body_idx];
		Scalar4 ex_space, ey_space, ez_space;

		exyzFromQuaternion(quat, ex_space, ey_space, ez_space);

		int3 image = d_rigidimage[body_idx];
		Scalar4 com = d_rigidcom[body_idx];

    // read particle position & write position
    for (unsigned int start = 0; start < nmax; start += blockDim.x){
      if (group_idx + start < nmax){
        unsigned int local_idx = body_idx * nmax + group_idx + start;
        idx = d_rigidparticleindices[local_idx];
        if (idx != INVALID_INDEX){

          Scalar4 posi = d_rigidparticlepos[local_idx]; // Local position (COM frame) of single particle
          Scalar3 ri; // Local position (space frame, relative to COM) of single particle
          Scalar3 abs_pos; // Global position that will be written to d_pos

          ri.x = ex_space.x * posi.x + ey_space.x * posi.y + ez_space.x * posi.z;
          ri.y = ex_space.y * posi.x + ey_space.y * posi.y + ez_space.y * posi.z;
          ri.z = ex_space.z * posi.x + ey_space.z * posi.y + ez_space.z * posi.z;

          abs_pos.x = com.x + ri.x;
          abs_pos.y = com.y + ri.y;
          abs_pos.z = com.z + ri.z;

          box.wrap(abs_pos, image);
          d_pos[idx] = make_scalar4(abs_pos.x, abs_pos.y, abs_pos.z, d_pos[idx].w);
          d_image[idx] = image;
        }
      }
    }
	}

}


//! kernel function to make displacement for each particle to each rigid body
__global__
void gpu_stokesrigid_rigiddis_kernel(Scalar4 *d_rigidcom,
				Scalar4 *d_rigidvel,
				Scalar4 *d_rigidorientation,
				Scalar4 *d_rigidangvel,
				unsigned int *d_rigidindex,
				unsigned int n_bodies,
				int3 *d_rigidimage,
				const BoxDim box,
				Scalar deltaT)
{
	unsigned int body_group_idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (body_group_idx < n_bodies){
		unsigned int body_idx = d_rigidindex[body_group_idx];
		Scalar4 com = d_rigidcom[body_idx];
		Scalar4 vel = d_rigidvel[body_idx];
		Scalar3 pos = make_scalar3(com.x, com.y, com.z) + make_scalar3(vel.x, vel.y, vel.z) * deltaT;
		int3 body_image = d_rigidimage[body_idx];

		box.wrap(pos, body_image);

		d_rigidcom[body_idx] = make_scalar4(pos.x, pos.y, pos.z, com.w);
		d_rigidimage[body_idx] = body_image;

		// Start: Advance the Quaternion
		Scalar4 quat = d_rigidorientation[body_idx];
		Scalar4 angvel = d_rigidangvel[body_idx];
		Scalar4 omegaq;

		vecquat(angvel, quat, omegaq);

		quat.x = quat.x + omegaq.x * deltaT * 0.5; // dq/dt = 1/2 w q. This is still one step integration!
		quat.y = quat.y + omegaq.y * deltaT * 0.5;
		quat.z = quat.z + omegaq.z * deltaT * 0.5;
		quat.w = quat.w + omegaq.w * deltaT * 0.5;
		normalize(quat);

		d_rigidorientation[body_idx] = quat;
		// End: Advance the Quaternion

	}

}


/*!
  	Wrap all the functions to compute hydrodynamic stress from force
	  Drive GPU kernel functions
	  S = Sum (K * F)
 */
void gpu_stokesrigid_StressFromForce_wrap(Scalar4 *d_force,
				Scalar3 *d_vecSii,
			        Scalar3 *d_vecSij,
				unsigned int *d_rigidindex,
				Scalar4 *d_rigidorientation,
				unsigned int n_bodies,
				unsigned int nmax,
				unsigned int *d_rigidparticleindices,
				Scalar4 *d_rigidparticlepos,
				dim3 grid_rigid_particle,
    				dim3 threads_rigid_particle,
				unsigned int rigidThreadDx,
			        unsigned int rigidThreadDy,
				Scalar3 *Sii_total,
			        Scalar3 *Sij_total)
{
  gpu_stokesrigid_rigidsigmaStress_kernel<<<grid_rigid_particle,threads_rigid_particle,rigidThreadDx*rigidThreadDy*sizeof(Scalar3)*2>>>(d_force,
    d_vecSii,
    d_vecSij,
    d_rigidindex,
    d_rigidorientation,
    n_bodies,
    nmax,
    d_rigidparticleindices,
    d_rigidparticlepos);

  unsigned int thread_for_dot = 512; // Must be 2^n
  unsigned int grid_for_dot = (n_bodies/thread_for_dot) + 1;

  // Temp var for vector sum.
  Scalar3 *vector_sum;
  cudaMalloc( (void**)&vector_sum, grid_for_dot*sizeof(Scalar3) );

  gpu_stokesrigid_VectorSumStepOne_kernel<<< grid_for_dot, thread_for_dot, thread_for_dot*sizeof(Scalar3) >>>(d_vecSii, vector_sum, n_bodies, d_rigidindex);
  gpu_stokesrigid_VectorSumStepTwo_kernel<<< 1, thread_for_dot, thread_for_dot*sizeof(Scalar3) >>>(vector_sum, grid_for_dot);

  cudaMemcpy(Sii_total, vector_sum, sizeof(Scalar3), cudaMemcpyDeviceToHost);

  gpu_stokesrigid_VectorSumStepOne_kernel<<< grid_for_dot, thread_for_dot, thread_for_dot*sizeof(Scalar3) >>>(d_vecSij, vector_sum, n_bodies, d_rigidindex);
  gpu_stokesrigid_VectorSumStepTwo_kernel<<< 1, thread_for_dot, thread_for_dot*sizeof(Scalar3) >>>(vector_sum, grid_for_dot);

  cudaMemcpy(Sij_total, vector_sum, sizeof(Scalar3), cudaMemcpyDeviceToHost);

  cudaFree(vector_sum);
}


__global__
void gpu_stokesrigid_xFcalculation_kernel(Scalar4 *d_net_force,
				Scalar3 *d_vecSii,
			        Scalar3 *d_vecSij,
				unsigned int *d_rigidindex,
				Scalar4 *d_rigidorientation,
				unsigned int n_bodies,
				unsigned int nmax,
				unsigned int *d_rigidparticleindices,
				Scalar4 *d_rigidparticlepos,
				Scalar *d_net_virial,
				unsigned int virial_pitch)
{
	unsigned int body_group_idx = blockIdx.x * blockDim.y + threadIdx.y;
	unsigned int group_idx = threadIdx.x;
	unsigned int body_idx;
	unsigned int idx;

	Scalar4 forcei, posi; // Force of single particle; local position (COM frame) of single particle
	Scalar3 ri; // local position (space frame, relative to COM) of single particle
	Scalar3 temp_Sii = make_scalar3(0.0,0.0,0.0); // Temp vector for stress calculating
	Scalar3 temp_Sij = make_scalar3(0.0,0.0,0.0); // Temp vector for stress calculating
	Scalar4 ex_space, ey_space, ez_space;

	if (body_group_idx < n_bodies){
		body_idx = d_rigidindex[body_group_idx];
		Scalar4 quat = d_rigidorientation[body_idx];

		exyzFromQuaternion(quat, ex_space, ey_space, ez_space);

		// read particle force
		for (unsigned int start = 0; start < nmax; start += blockDim.x){
			if (group_idx + start < nmax){
				unsigned int local_idx = body_idx * nmax + group_idx + start;
				idx = d_rigidparticleindices[local_idx];
				if (idx != INVALID_INDEX){
					forcei = d_net_force[idx];
					posi = d_rigidparticlepos[local_idx];
					ri.x = ex_space.x * posi.x + ey_space.x * posi.y + ez_space.x * posi.z;
                			ri.y = ex_space.y * posi.x + ey_space.y * posi.y + ez_space.y * posi.z;
                			ri.z = ex_space.z * posi.x + ey_space.z * posi.y + ez_space.z * posi.z;

					temp_Sii += make_scalar3(d_net_virial[0*virial_pitch+idx] - ri.x * forcei.x, d_net_virial[3*virial_pitch+idx] - ri.y * forcei.y, d_net_virial[5*virial_pitch+idx] - ri.z * forcei.z);
					temp_Sij += (make_scalar3(d_net_virial[1*virial_pitch+idx], d_net_virial[2*virial_pitch+idx], d_net_virial[4*virial_pitch+idx]) - 0.5 * make_scalar3(ri.y * forcei.x + ri.x * forcei.y, ri.z * forcei.x + ri.x * forcei.z, ri.z * forcei.y + ri.y * forcei.z));
					// x_p * F_p - (x_p-x_R) * F_p
				}
			}
		}
	}

	// Save them into shared memory
	shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y] = temp_Sii;
	shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y + blockDim.x * blockDim.y] = temp_Sij;

	// Reduction with each rigid body
	// Note: It seems that exchanging threadIdx.x and threadIdx.y can be faster. (Not priority since not the bottleneck)
	int offs = blockDim.x >> 1;
	while (offs > 0)
       	{
		__syncthreads();
            	if (threadIdx.x < offs)
                {
                	shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y] += shared_rigid_vel[threadIdx.x + offs + blockDim.x * threadIdx.y];
			shared_rigid_vel[threadIdx.x + blockDim.x * threadIdx.y + blockDim.x * blockDim.y] += shared_rigid_vel[threadIdx.x + offs + blockDim.x * threadIdx.y + blockDim.x * blockDim.y];
                }
            	offs >>= 1;
        }
	__syncthreads();
        if (threadIdx.x == 0)
	{
		if (body_group_idx < n_bodies){
			d_vecSii[body_idx] = shared_rigid_vel[blockDim.x * threadIdx.y];
			d_vecSij[body_idx] = shared_rigid_vel[blockDim.x * threadIdx.y + blockDim.x * blockDim.y];
		}
	}
	// Save the stress to global memory
}


// Calculate xF from x and F
void gpu_stokesrigid_xFcalculation_wrap(Scalar4 *d_net_force,
				Scalar3 *d_vecSii,
			        Scalar3 *d_vecSij,
				unsigned int *d_rigidindex,
				Scalar4 *d_rigidorientation,
				unsigned int n_bodies,
				unsigned int nmax,
				unsigned int *d_rigidparticleindices,
				Scalar4 *d_rigidparticlepos,
				dim3 grid_rigid_particle,
    				dim3 threads_rigid_particle,
				unsigned int rigidThreadDx,
			        unsigned int rigidThreadDy,
				Scalar3 *Sii_total,
			        Scalar3 *Sij_total,
				Scalar *d_net_virial,
				unsigned int virial_pitch){

	gpu_stokesrigid_xFcalculation_kernel<<<grid_rigid_particle,threads_rigid_particle,rigidThreadDx*rigidThreadDy*sizeof(Scalar3)*2>>>(d_net_force,
				d_vecSii,
			        d_vecSij,
				d_rigidindex,
				d_rigidorientation,
				n_bodies,
				nmax,
				d_rigidparticleindices,
				d_rigidparticlepos,
				d_net_virial,
				virial_pitch);
	// Sum over the x_R*F_p terms


	// Then, sum over all the rigid bodies
	// Maybe this can be simplified, but just in case the stress can be used somewhere else

	unsigned int thread_for_dot = 512; // Must be 2^n
	unsigned int grid_for_dot = (n_bodies/thread_for_dot) + 1;

	// Temp var for vector sum.
	Scalar3 *vector_sum;
	cudaMalloc( (void**)&vector_sum, grid_for_dot*sizeof(Scalar3) );

	gpu_stokesrigid_VectorSumStepOne_kernel<<< grid_for_dot, thread_for_dot, thread_for_dot*sizeof(Scalar3) >>>(d_vecSii, vector_sum, n_bodies, d_rigidindex);
	gpu_stokesrigid_VectorSumStepTwo_kernel<<< 1, thread_for_dot, thread_for_dot*sizeof(Scalar3) >>>(vector_sum, grid_for_dot);

	cudaMemcpy(Sii_total, vector_sum, sizeof(Scalar3), cudaMemcpyDeviceToHost);

	gpu_stokesrigid_VectorSumStepOne_kernel<<< grid_for_dot, thread_for_dot, thread_for_dot*sizeof(Scalar3) >>>(d_vecSij, vector_sum, n_bodies, d_rigidindex);
	gpu_stokesrigid_VectorSumStepTwo_kernel<<< 1, thread_for_dot, thread_for_dot*sizeof(Scalar3) >>>(vector_sum, grid_for_dot);

	cudaMemcpy(Sij_total, vector_sum, sizeof(Scalar3), cudaMemcpyDeviceToHost);

	cudaFree(vector_sum);

}


//! kernel function to get velocity for each particle from each rigid body with a uniform strain
// Used for shear simulation
__global__
void gpu_stokesrigid_UniformStrain_kernel(Scalar4 *d_vel,
			        Scalar Exy,
				unsigned int *d_rigidindex,
				Scalar4 *d_rigidorientation,
				unsigned int n_bodies,
				unsigned int nmax,
				unsigned int *d_rigidparticleindices,
				Scalar4 *d_rigidparticlepos)
{
	unsigned int body_group_idx = blockIdx.x * blockDim.y + threadIdx.y;
	unsigned int group_idx = threadIdx.x;
	unsigned int body_idx;
	unsigned int idx;

	if (body_group_idx < n_bodies){
		body_idx = d_rigidindex[body_group_idx];

		Scalar4 quat = d_rigidorientation[body_idx];
		Scalar4 ex_space, ey_space, ez_space;

		exyzFromQuaternion(quat, ex_space, ey_space, ez_space);

		// read particle position & write velocity
		for (unsigned int start = 0; start < nmax; start += blockDim.x){
			if (group_idx + start < nmax){
				unsigned int local_idx = body_idx * nmax + group_idx + start;
				idx = d_rigidparticleindices[local_idx];
				if (idx != INVALID_INDEX){

					Scalar4 posi = d_rigidparticlepos[local_idx]; // Local position (COM frame) of single particle

					// Local position (space frame, relative to COM) of single particle
					Scalar rix = ex_space.x * posi.x + ey_space.x * posi.y + ez_space.x * posi.z;
                			Scalar riy = ex_space.y * posi.x + ey_space.y * posi.y + ez_space.y * posi.z;

					d_vel[idx] = make_scalar4(Exy * riy, Exy * rix, 0.0, 0.0);
				}
			}
		}
	}
}


//! kernel function to make displacement for each particle to each rigid body, version with shear rate
__global__
void gpu_stokesrigid_rigiddisShear_kernel(Scalar4 *d_rigidcom,
				Scalar4 *d_rigidvel,
				Scalar4 *d_rigidorientation,
				Scalar4 *d_rigidangvel,
				unsigned int *d_rigidindex,
				unsigned int n_bodies,
				int3 *d_rigidimage,
				const BoxDim box,
				Scalar deltaT,
				Scalar Exy)
{
	unsigned int body_group_idx = blockIdx.x * blockDim.x + threadIdx.x;
	if (body_group_idx < n_bodies){
		unsigned int body_idx = d_rigidindex[body_group_idx];
		Scalar4 com = d_rigidcom[body_idx];
		Scalar4 vel = d_rigidvel[body_idx];

		vel.x += 2.0 * Exy * com.y;

		Scalar3 pos = make_scalar3(com.x, com.y, com.z) + make_scalar3(vel.x, vel.y, vel.z) * deltaT;
		int3 body_image = d_rigidimage[body_idx];

		box.wrap(pos, body_image);

		d_rigidcom[body_idx] = make_scalar4(pos.x, pos.y, pos.z, com.w);
		d_rigidimage[body_idx] = body_image;

		// Start: Advance the Quaternion
		Scalar4 quat = d_rigidorientation[body_idx];
		Scalar4 angvel = d_rigidangvel[body_idx];

		angvel.z -= Exy;

		Scalar4 omegaq;

		vecquat(angvel, quat, omegaq);

		quat.x = quat.x + omegaq.x * deltaT * 0.5; // dq/dt = 1/2 w q. This is still one step integration!
		quat.y = quat.y + omegaq.y * deltaT * 0.5;
		quat.z = quat.z + omegaq.z * deltaT * 0.5;
		quat.w = quat.w + omegaq.w * deltaT * 0.5;
		normalize(quat);

		d_rigidorientation[body_idx] = quat;
		// End: Advance the Quaternion

	}

}


/*! \param d_pos array of particle positions
    \param d_vel array of particle velocities
    \param d_group_members Device array listing the indicies of the mebers of the group to integrate
    \param group_size Number of members in the group
    \param box Box dimensions for periodic boundary condition handling
    \param deltaT timestep
    \param block_size optimum block size returned by an autotuner
    \param d_net_force net force on the particles
    \param T temperature
    \param timestep time step
    \param seed seed
    \param alpha splitting coefficient for Ewald summation
    \param ewald_cut cut off radius for Ewald summation
    \param ewald_dr delta r for table look up
    \param ewald_n number for table look up
    \param d_ewaldC Ewald coefficients for real space sum
    \param d_gridk parameters for Ewald reciprocal space sum
    \param d_gridX grid for Fx
    \param d_gridY grid for Fy
    \param d_gridZ grid for Fz
    \param plan cudaFFT plan
    \param Nx Nx
    \param Ny Ny
    \param Nz Nz
    \param d_n_neigh Number of neighbors for every particle
    \param d_nlist Neighbor list of every particle, 2D array, can be accessed by --->
    \param nli Index lookup helper for d_nlist
    \param cheb_an Cheb coefficients
    \param n_cheb order of Cheb approximation
    \param N_total total number of particles, not sure if it is needed
    \param Lagr Order of Lagrangian interpolation
    \param gridL size of the grids
    \param d_diameter diameters of the particles

*/
cudaError_t gpu_stokesrigid_step_one(Scalar4 *d_pos,
  Scalar4 *d_vel,
  unsigned int *d_group_members,
  unsigned int group_size,
  const BoxDim& box,
  Scalar deltaT,
  Scalar4 *d_net_force,
  const Scalar T,
  const unsigned int timestep,
  const unsigned int seed,
  Scalar xi,
  Scalar eta,
  Scalar self,
  const int P,
  Scalar3 gridh,
  Scalar ewald_cut,
  Scalar ewald_dr,
  int ewald_n,
  Scalar cheb_error,
  Scalar4 *d_ewaldC1,
  int& Lanczos,
  Scalar4 *d_gridk,
  CUFFTCOMPLEX *d_gridX,
  CUFFTCOMPLEX *d_gridY,
  CUFFTCOMPLEX *d_gridZ,
  cufftHandle plan,
  const int Nx,
  const int Ny,
  const int Nz,
  const unsigned int *d_n_neigh,
  const unsigned int *d_nlist,
  const unsigned int *d_headlist,
  const unsigned int N_total,
  const Scalar *d_diameter,
  Scalar4 *d_rigidvel,
  Scalar4 *d_rigidangvel,
  //Scalar4 *d_rigidforce,
  //Scalar4 *d_rigidtorque,
  unsigned int *d_rigidindex,
  Scalar4 *d_rigidorientation,
  unsigned int n_bodies,
  unsigned int nmax,
  unsigned int *d_rigidparticleindices,
  Scalar4 *d_rigidparticlepos,
  Scalar4 *d_rigidcom,
  int3 *d_rigidimage,
  unsigned int rigidThreadDx,
  unsigned int rigidThreadDy,
  unsigned int rigidGrid,
  unsigned int N_rigid_total,
  Scalar *d_rigidbodymass,
  Scalar4 *d_rigidmomentinertia,
  int3 *d_image,
  bool hydrostress_now,
  Scalar3 *d_vecSii,
  Scalar3 *d_vecSij,
  Scalar3 *SBii_total,
  Scalar3 *SBij_total,
  Scalar3 *SPii_total,
  Scalar3 *SPij_total,
  Scalar *d_net_virial,
  unsigned int virial_pitch,
  Scalar shear_rate,
  Scalar3 *test)
    {

      unsigned int block_size = 256; // For particle calculation

      unsigned int NxNyNz = Nx*Ny*Nz;
      // setup the grid to run the kernel
      // block for particle calculation
      dim3 grid( (group_size/block_size) + 1, 1, 1);
      dim3 threads(block_size, 1, 1);

      // block for grid calculation
      int gridBlockSize = ( NxNyNz > block_size ) ? block_size : NxNyNz;
      int gridNBlock = ( NxNyNz + gridBlockSize - 1 ) / gridBlockSize ;

      // block for rigid body (each particle) calculation
      dim3 grid_rigid_particle( rigidGrid, 1, 1);
      dim3 threads_rigid_particle(rigidThreadDx, rigidThreadDy, 1);

      // block for rigid body (each rigid body) calculation
      unsigned int thread_rigid = 512;
      unsigned int grid_rigid = n_bodies/thread_rigid+1;

      // Bind texture memory
      tables1_tex.normalized = false; // Not normalized
      tables1_tex.filterMode = cudaFilterModePoint; // Filter mode: floor of the index
      // One dimension, Read mode: ElementType(Get what we write)
      cudaBindTexture(0, tables1_tex, d_ewaldC1, sizeof(Scalar4) * (ewald_n+1));

      // Same for the positions and forces
      pos_tex.normalized = false; // Not normalized
      pos_tex.filterMode = cudaFilterModePoint; // Filter mode: floor of the index
      cudaBindTexture(0, pos_tex, d_pos, sizeof(Scalar4) * N_total);

      gpu_stokes_SetGridk_kernel<<<gridNBlock,gridBlockSize>>>(d_gridk,Nx,Ny,Nz,NxNyNz,box,xi,eta);

      // Create the matrix wrappers
      cuspRigidSaddle MSS0_Matrix(d_vel,
            d_pos,
            d_group_members,
            d_ewaldC1,
            d_gridk,
            d_gridX,
            d_gridY,
            d_gridZ,
            d_n_neigh,
            d_nlist,
            d_headlist,
            d_diameter,
            d_rigidindex,
            d_rigidorientation,
            d_rigidparticleindices,
            d_rigidparticlepos,
            group_size,
            box,
            xi,
            eta,
            self,
            P,
            gridh,
            ewald_cut,
            ewald_dr,
            ewald_n,
            plan,
            Nx,
            Ny,
            Nz,
            NxNyNz,
            N_total,
            n_bodies,
            nmax,
            N_rigid_total,
            grid,
            threads,
            gridBlockSize,
            gridNBlock,
            grid_rigid_particle,
            threads_rigid_particle,
            rigidThreadDx,
            rigidThreadDy);
            // Set up the preconditioner

      Scalar4 *d_preconditioner_internal; // Internal arrays for preconditioning
      cudaMalloc( (void**)&d_preconditioner_internal, (N_total + 2*N_rigid_total)*sizeof(Scalar4));
      Scalar radius_rep;
      cudaMemcpy(&radius_rep, d_diameter, sizeof(Scalar), cudaMemcpyDeviceToHost);
      radius_rep = radius_rep / 2.0 ;

      cuspPreCondSigma PreCond_Matrix(d_preconditioner_internal,
        d_group_members,
        d_rigidindex,
        d_rigidorientation,
        d_rigidparticleindices,
        d_rigidparticlepos,
        d_rigidbodymass,
        d_rigidmomentinertia,
        group_size,
        N_total,
        n_bodies,
        nmax,
        N_rigid_total,
        grid,
        threads,
        grid_rigid_particle,
        threads_rigid_particle,
        rigidThreadDx,
        rigidThreadDy,
        grid_rigid,
        thread_rigid,
        radius_rep);

      // Allocation of memory for the vectors
      cusp::array1d<Scalar, cusp::device_memory> cuspX1(4*N_total+8*N_rigid_total, Scalar(0));
      cusp::array1d<Scalar, cusp::device_memory> cuspB1(4*N_total+8*N_rigid_total, Scalar(0));
      cusp::array1d<Scalar, cusp::device_memory> cuspB_Brownian(4*N_total+8*N_rigid_total, Scalar(0));

      // Get raw pointers
      Scalar4 *X1_ptr = (Scalar4 *) thrust::raw_pointer_cast(&cuspX1[0]);
      Scalar4 *B1_ptr = (Scalar4 *) thrust::raw_pointer_cast(&cuspB1[0]);
      Scalar4 *B_Brownian_ptr = (Scalar4 *) thrust::raw_pointer_cast(&cuspB_Brownian[0]);
      // Rigid Index: N_total, N_total + N_rigid_total

      // Get force and torque for rigid body from single particle force
      gpu_stokesrigid_rigidsigma_kernel<<<grid_rigid_particle,threads_rigid_particle,rigidThreadDx*rigidThreadDy*sizeof(Scalar3)*2>>>(d_net_force,
        &B1_ptr[N_total],
        &B1_ptr[N_total+N_rigid_total],
        d_rigidindex,
        d_rigidorientation,
        n_bodies,
        nmax,
        d_rigidparticleindices,
        d_rigidparticlepos);

      Scalar3 SPxFii, SPxFij, SPsuii, SPsuij; // Two parts of Particle force stress

      if (hydrostress_now){
        // First, calculate SPxF contribution
        gpu_stokesrigid_xFcalculation_wrap(d_net_force,
          d_vecSii,
          d_vecSij,
          d_rigidindex,
          d_rigidorientation,
          n_bodies,
          nmax,
          d_rigidparticleindices,
          d_rigidparticlepos,
          grid_rigid_particle,
          threads_rigid_particle,
          rigidThreadDx,
          rigidThreadDy,
          &SPxFii,
          &SPxFij,
          d_net_virial,
          virial_pitch);
        //printf("SPxFterm: %.8f, %.8f, %.8f, %.8f, %.8f, %.8f\n", SPxFii.x, SPxFii.y, SPxFii.z, SPxFij.x, SPxFij.y, SPxFij.z);
      }

      Scalar NegExy = -shear_rate/2; // RHS is negative K'*E..
      // Strain
      if (shear_rate!=0.0){

        gpu_stokesrigid_UniformStrain_kernel<<<grid_rigid_particle,threads_rigid_particle>>>(B1_ptr,
          NegExy,
          d_rigidindex,
          d_rigidorientation,
          n_bodies,
          nmax,
          d_rigidparticleindices,
          d_rigidparticlepos);

      }

      if (T>0.0){
        // Get bd velocity first
        gpu_stokes_CombinedMobilityBrownian_wrap(d_pos,
          d_group_members,
          group_size,
          box,
          deltaT,
          B_Brownian_ptr, // Fill in the B1 vector, -uB is supposed but uB is used. However, this shouldn't change the nature
          T,
          timestep,
          seed,
          xi,
          eta,
          P,
          ewald_cut,
          ewald_dr,
          ewald_n,
          d_ewaldC1,
          d_gridk,
          d_gridX,
          d_gridY,
          d_gridZ,
          plan,
          Nx,
          Ny,
          Nz,
          d_n_neigh,
          d_nlist,
          d_headlist,
          Lanczos,
          N_total,
          NxNyNz,
          grid,
          threads,
          gridBlockSize,
          gridNBlock,
          gridh,
          d_diameter,
          cheb_error,
          self);
      }

      if (!hydrostress_now){
        // If not calculating stress at the current step
        // Combine the RHS's to save one solve
        gpu_stokes_LinearCombination_kernel<<<grid, threads>>>(B1_ptr, B_Brownian_ptr, B1_ptr, 1.0, 1.0, group_size, d_group_members);
      }

      // Get velocity and angular velocity for rigid body by GMRES
      cusp::convergence_monitor<Scalar> monitor1(cuspB1, GMRES_ITER_NUM, KRYLOV_ERROR);
      cusp::krylov::gmres( MSS0_Matrix, cuspX1, cuspB1, GMRES_RESET, monitor1, PreCond_Matrix );
      //monitor1.print();

      if (hydrostress_now){
        // Now calculate the SU contribution of S_P
        gpu_stokesrigid_StressFromForce_wrap(X1_ptr,
          d_vecSii,
          d_vecSij,
          d_rigidindex,
          d_rigidorientation,
          n_bodies,
          nmax,
          d_rigidparticleindices,
          d_rigidparticlepos,
          grid_rigid_particle,
          threads_rigid_particle,
          rigidThreadDx,
          rigidThreadDy,
          &SPsuii,
          &SPsuij);
        //printf("SPSUterm: %.8f, %.8f, %.8f, %.8f, %.8f, %.8f\n", SPsuii.x, SPsuii.y, SPsuii.z, SPsuij.x, SPsuij.y, SPsuij.z);

        if (T>0.0){
          // If Stress is calculated at the step AND T > 0.0
          // We have to calculate Brownian velocity separately

          gpu_stokesrigid_SetEqual_kernel<<<grid_rigid, thread_rigid>>>(d_rigidvel, &X1_ptr[N_total], n_bodies, d_rigidindex);
          gpu_stokesrigid_SetEqual_kernel<<<grid_rigid, thread_rigid>>>(d_rigidangvel, &X1_ptr[N_total+N_rigid_total], n_bodies, d_rigidindex);
          // Save the deterministic result first
          thrust::fill( cuspX1.begin(), cuspX1.end(), 0 );

          cusp::convergence_monitor<Scalar> monitor2(cuspB_Brownian, GMRES_ITER_NUM, KRYLOV_ERROR);
          cusp::krylov::gmres( MSS0_Matrix, cuspX1, cuspB_Brownian, GMRES_RESET, monitor2, PreCond_Matrix );
          //monitor2.print();

          gpu_stokes_LinearCombination_kernel<<<grid_rigid, thread_rigid>>>(d_rigidvel, &X1_ptr[N_total], &X1_ptr[N_total], 1.0, 1.0, n_bodies, d_rigidindex);
          gpu_stokes_LinearCombination_kernel<<<grid_rigid, thread_rigid>>>(d_rigidangvel, &X1_ptr[N_total+N_rigid_total], &X1_ptr[N_total+N_rigid_total], 1.0, 1.0, n_bodies, d_rigidindex);
        }
      }

      // Start:test of divM, comment
      // Move the rigid bodies, need a new kernel function.
      gpu_stokesrigid_rigiddisShear_kernel<<<grid_rigid,thread_rigid>>>(d_rigidcom, &X1_ptr[N_total], d_rigidorientation, &X1_ptr[N_total+N_rigid_total], d_rigidindex, n_bodies, d_rigidimage, box, -deltaT, NegExy);
      // Update d_pos: This is important. We have to upgrade mobility tensor.
      gpu_stokesrigid_setdpos_kernel<<<grid_rigid_particle,threads_rigid_particle>>>(d_pos, d_image, d_rigidindex, d_rigidorientation, n_bodies, nmax, d_rigidparticleindices, d_rigidparticlepos, d_rigidcom, d_rigidimage, box);
      // End: test of divM, comment

      Scalar3 SB0ii, SB0ij, SB1ii, SB1ij; // Two parts of Bronwian stress

      if (T>0.0){

        // Random finite difference method based on Delong et al., 2015
        // Implemented 2017, Jan.

        // Start Random Finite Difference to compute the Brownian Drift
        thrust::fill( cuspB1.begin(), cuspB1.end(), 0 );
        thrust::fill( cuspX1.begin(), cuspX1.end(), 0 );

        // Generate Random Psi
        gpu_stokes_BrownianGenerate_kernel<<<grid_rigid,thread_rigid>>>( &B1_ptr[N_total], n_bodies, d_rigidindex, timestep * 0x13579913, seed );
        gpu_stokes_BrownianGenerate_kernel<<<grid_rigid,thread_rigid>>>( &B1_ptr[N_total+N_rigid_total], n_bodies, d_rigidindex, timestep * 0x13579913 * 0x19367281, seed );
        // Hopefully this can give us independent random variables

        // Calculate the velocity after a partial timestep.
        cusp::convergence_monitor<Scalar> monitor3(cuspB1, GMRES_ITER_NUM, KRYLOV_ERROR);
        cusp::krylov::gmres( MSS0_Matrix, cuspX1, cuspB1, GMRES_RESET, monitor3, PreCond_Matrix );
        //monitor3.print();

        // Save the velocity/angular velocity of first step
        gpu_stokesrigid_SetEqual_kernel<<<grid_rigid, thread_rigid>>>(d_rigidvel, &X1_ptr[N_total], n_bodies, d_rigidindex);
        gpu_stokesrigid_SetEqual_kernel<<<grid_rigid, thread_rigid>>>(d_rigidangvel, &X1_ptr[N_total+N_rigid_total], n_bodies, d_rigidindex);
        // Note: still negative velocity!

        if (hydrostress_now){
          gpu_stokesrigid_StressFromForce_wrap(X1_ptr,
            d_vecSii,
            d_vecSij,
            d_rigidindex,
            d_rigidorientation,
            n_bodies,
            nmax,
            d_rigidparticleindices,
            d_rigidparticlepos,
            grid_rigid_particle,
            threads_rigid_particle,
            rigidThreadDx,
            rigidThreadDy,
            &SB0ii,
            &SB0ij);
          //printf("SB0term: %.8f, %.8f, %.8f, %.8f, %.8f, %.8f\n", SB0ii.x, SB0ii.y, SB0ii.z, SB0ij.x, SB0ij.y, SB0ij.z);
        }

        gpu_stokesrigid_rigiddis_kernel<<<grid_rigid,thread_rigid>>>(d_rigidcom, &B1_ptr[N_total], d_rigidorientation, &B1_ptr[N_total+N_rigid_total], d_rigidindex, n_bodies, d_rigidimage, box, RFD_DELTA);
        gpu_stokesrigid_setdpos_kernel<<<grid_rigid_particle,threads_rigid_particle>>>(d_pos, d_image, d_rigidindex, d_rigidorientation, n_bodies, nmax, d_rigidparticleindices, d_rigidparticlepos, d_rigidcom, d_rigidimage, box);

        // Calculate the velocity after a partial timestep.
        cusp::convergence_monitor<Scalar> monitor4(cuspB1, GMRES_ITER_NUM, KRYLOV_ERROR);
        cusp::krylov::gmres( MSS0_Matrix, cuspX1, cuspB1, GMRES_RESET, monitor4, PreCond_Matrix );
        //monitor4.print();

        if (hydrostress_now){
          gpu_stokesrigid_StressFromForce_wrap(X1_ptr,
            d_vecSii,
            d_vecSij,
            d_rigidindex,
            d_rigidorientation,
            n_bodies,
            nmax,
            d_rigidparticleindices,
            d_rigidparticlepos,
            grid_rigid_particle,
            threads_rigid_particle,
            rigidThreadDx,
            rigidThreadDy,
            &SB1ii,
            &SB1ij);
          //printf("SB1term: %.8f, %.8f, %.8f, %.8f, %.8f, %.8f\n", SB1ii.x, SB1ii.y, SB1ii.z, SB1ij.x, SB1ij.y, SB1ij.z);
        }

        gpu_stokesrigid_rigiddis_kernel<<<grid_rigid,thread_rigid>>>(d_rigidcom, &B1_ptr[N_total], d_rigidorientation, &B1_ptr[N_total+N_rigid_total], d_rigidindex, n_bodies, d_rigidimage, box, -RFD_DELTA);
        gpu_stokes_LinearCombination_kernel<<<grid_rigid, thread_rigid>>>(d_rigidvel, &X1_ptr[N_total], d_rigidvel, -1.0, 1.0, n_bodies, d_rigidindex);
        gpu_stokes_LinearCombination_kernel<<<grid_rigid, thread_rigid>>>(d_rigidangvel, &X1_ptr[N_total+N_rigid_total], d_rigidangvel, -1.0, 1.0, n_bodies, d_rigidindex);

        // Start: test of divM
/*
        Scalar4 rigidvel0, rigidvel1, rigidvel01, rigidvel11;
        cudaMemcpy(&rigidvel0, &d_rigidvel[0], sizeof(Scalar4), cudaMemcpyDeviceToHost);
        cudaMemcpy(&rigidvel1, &d_rigidvel[1], sizeof(Scalar4), cudaMemcpyDeviceToHost);
        cudaMemcpy(&rigidvel01, &d_rigidangvel[0], sizeof(Scalar4), cudaMemcpyDeviceToHost);
        cudaMemcpy(&rigidvel11, &d_rigidangvel[1], sizeof(Scalar4), cudaMemcpyDeviceToHost);
        test[0] = make_scalar3(rigidvel0.x,rigidvel0.y,rigidvel0.z)/RFD_DELTA*(-T);
        test[1] = make_scalar3(rigidvel1.x,rigidvel1.y,rigidvel1.z)/RFD_DELTA*(-T);
        test[2] = make_scalar3(rigidvel01.x,rigidvel01.y,rigidvel01.z)/RFD_DELTA*(-T);
        test[3] = make_scalar3(rigidvel11.x,rigidvel11.y,rigidvel11.z)/RFD_DELTA*(-T);
*/
        // End: test of divM

        // Move the rigid body by changing the COM position and quaternion
        // Start: test of divM, comment
        gpu_stokesrigid_rigiddis_kernel<<<grid_rigid,thread_rigid>>>(d_rigidcom, d_rigidvel, d_rigidorientation, d_rigidangvel, d_rigidindex, n_bodies, d_rigidimage, box, -deltaT/RFD_DELTA*T);
        // End: test of divM, comment
      }

      if (hydrostress_now){
        // Finally, add up all the contributions:
        if (T>0.0){
          *SBii_total = (SB1ii - SB0ii)/RFD_DELTA*T;
          *SBij_total = (SB1ij - SB0ij)/RFD_DELTA*T;
        }
        *SPii_total = SPxFii + SPsuii;
        *SPij_total = SPxFij + SPsuij;
      }

      cudaUnbindTexture(tables1_tex);
      cudaUnbindTexture(pos_tex);

      cudaFree(d_preconditioner_internal);

      return cudaSuccess;
    }
