// Maintainer: joaander
// Modified by Gang Wang

#include "hoomd/hoomd_config.h"
#include "IntegrationMethodTwoStep.h"
#include "Variant.h"
#include <cufft.h>
#include "RigidData.h"
#include "RigidBodyGroup.h"
#include "ShearFunction.h"
#include "NeighborList.h"

#ifndef SINGLE_PRECISION
#define CUFFTCOMPLEX cufftComplex
#else
#define CUFFTCOMPLEX cufftComplex
#endif

#ifndef __STOKESRIGID_H__
#define __STOKESRIGID_H__

/*! \file StokesRigid.h
    \brief Declares the StokesRigid class with rigid body
*/

#ifdef NVCC
#error This header cannot be compiled by nvcc
#endif

//! Integrates the system forward considering hydrodynamic interactions by GPU
/*! Implements overdamped integration (one step) through IntegrationMethodTwoStep interface, runs on the GPU
*/
class StokesRigid : public IntegrationMethodTwoStep
    {
    public:
      //! Constructs the integration method and associates it with the system
      StokesRigid(boost::shared_ptr<SystemDefinition> sysdef,
        boost::shared_ptr<ParticleGroup> group,
        boost::shared_ptr<Variant> T,
        unsigned int seed,
        boost::shared_ptr<NeighborList> nlist,
        Scalar xi,
        Scalar error,
        bool hydrostress,
        std::string fname);

      virtual ~StokesRigid();

      //! Set a new temperature
      /*! \param T new temperature to set */
      void setT(boost::shared_ptr<Variant> T)
      {
        m_T = T;
      }

      //! Performs the first step of the integration
      virtual void integrateStepOne(unsigned int timestep);

      //! Performs the second step of the integration
      virtual void integrateStepTwo(unsigned int timestep);

      //! Set the parameters for Ewald summation
      void setParams();

      unsigned int nextPowerOf2(unsigned int n);

      //! Get the number of degrees of freedom granted to a given group
      virtual unsigned int getNDOF(boost::shared_ptr<ParticleGroup> query_group);

      //! Validate that all members in the particle group are valid (throw an exception if they are not)
      virtual void validateGroup();

      PDataFlags getRequestedPDataFlags();

      //! Sets the delimiter to use between fields
      void setDelimiter(const std::string& delimiter){
        m_delimiter = delimiter;
      }

      void setHydrostressPeriod(unsigned int hydrostress_period){
        m_hydrostress_period = hydrostress_period;
      }

      //! Set the shear rate and shear frequency
      void setShear(boost::shared_ptr<ShearFunction> shear_func, Scalar max_strain) {
          m_shear_func = shear_func;
          m_max_strain = max_strain;
      }

    protected:
      boost::shared_ptr<Variant> m_T;   //!< The Temperature of the Stochastic Bath
      unsigned int m_seed;              //!< The seed for the RNG of the Stochastic Bath

      cufftHandle plan;                    //!< Used for the Fast Fourier Transformations performed on the GPU

      boost::shared_ptr<NeighborList> m_nlist;    //!< The neighborlist to use for the computation

      Scalar m_ewald_cut;                           //!< Real space cutoff
      GPUArray<Scalar4> m_ewaldC1;                   //!< Real space Ewald coefficients table
      int m_ewald_n;                           //!< Number of entries in table of Ewald coefficients
      Scalar m_ewald_dr;                           //!< Real space Ewald table spacing

      int m_Nx;                                //!< Number of grid points in x direction
      int m_Ny;                                //!< Number of grid points in y direction
      int m_Nz;                                //!< Number of grid points in z direction

      GPUArray<Scalar4> m_gridk;      //!< k-vectors for each grid point
      GPUArray<CUFFTCOMPLEX> m_gridX;      //!< x component of the grid based force
      GPUArray<CUFFTCOMPLEX> m_gridY;      //!< x component of the grid based force
      GPUArray<CUFFTCOMPLEX> m_gridZ;      //!< x component of the grid based force

      // Rigid body part
      boost::shared_ptr<RigidData> m_rigid_data;  //!< Pointer to rigid data
      boost::shared_ptr<RigidBodyGroup> m_body_group; //!< Group of rigid bodies to work with
      unsigned int m_n_bodies;                    //!< Number of rigid bodies

      // For stress calculation
      GPUArray<Scalar3> m_vecSii; //!< Vector of stresslet, ii part, 11, 22, 33
      GPUArray<Scalar3> m_vecSij; //!< Vector of stresslet, ij part, 12, 13, 23

      Scalar m_max_strain; //!< Maximum total strain before box resizing
      boost::shared_ptr<ShearFunction> m_shear_func; //!< mutable shared pointer towards a ShearFunction object

      Scalar m_xi;                   //!< ewald splitting parameter xi
      Scalar m_self; //!< self piece
      Scalar m_gaussm;  //!< Gaussian width in standard deviations for wave space spreading/contraction
      int m_gaussP;     //!< Number of points in each dimension for Gaussian support
      Scalar m_eta;     //!< Gaussian spreading parameter
      Scalar3 m_gridh;  //!< Size of the grid box in 3 direction

      int m_m_Lanczos;       //!< Number of Lanczos Iterations to use for calculation of Brownian displacement

      Scalar m_error;  //!< Error tolerance for all calculations

    private:

      bool m_hydrostress; //!< Compute hydrodynamic stresslet or not

      //! The period of stress calculation
      unsigned int m_hydrostress_period;
      bool m_hydrostress_now; //!< Compute hydrodynamic stresslet or not for current timestep
      unsigned int m_hydrostress_lastcompute; //!< Last timestep we compute hydrodydamic stress

      //! The delimiter to put between columns in the file
      std::string m_delimiter;
      std::ofstream m_file;   //!< The file we write out to

      //! Helper function to write out the header
      void writeHeader();

    };

//! Exports the Stokes class to python
void export_StokesRigid();

#endif
