## \package pse_rigid.integrate
# class of integrator for rigid bodies with hydrodynamic interactions

import _pse_rigidv0

from hoomd_script.integrate import _integration_method
from hoomd_script import util
from hoomd_script import globals
import hoomd
from hoomd_script import variant
from hoomd_script import compute
from hoomd_script import pair
import math
import shear_function

## One step overdamped integration with hydrodynamic interactions <br>
#
# Contributors: <br>
# Rigid body part: Gang Wang <br>
# Mobility and Brownian calculation: Andrew Fiore <br>
# NOTE: To ensure the best preconditioner performance, it is suggested to set the mass of each bead to be 1.0. <br>
# Usage example with changing shear field:
# \code
# bd = pse_rigidv0.integrate.pse_rigid(group = all, T = TB, stress_filename = 'Pressure.txt', stress_period = 10)
# run(5000)
# steady_shear = pse_rigidv0.shear_function.steady(dt = mass, shear_rate = shear) # create a shear_function object
# bd.set_params(function_form = steady_shear) # The function_form can also be provided when initializing the object
# box_resize1 = update.box_resize(xy = pse_rigidv0.variant.shear_variant(function_form = steady_shear, total_timestep = 3000)) # Also need to create a box_resize object
# box_resize1.set_params(scale_particles=False) # DO NOT SCALE PARTICLES
# run(3000)
# box_resize1.disable() # Disable the box_resize before creating another one
# shear_function = pse_rigidv0.shear_function.chirp(dt = mass, amplitude = shear, omega_0 = freq_0, omega_f = freq_f, periodT = tf)
# bd.set_params(function_form = shear_function)
# box_resize = update.box_resize(xy = pse_rigidv0.variant.shear_variant(function_form = shear_function, total_timestep = iter))
# box_resize.set_params(scale_particles=False)
# \endcode
class pse_rigid(_integration_method):
    ## Specifies the pse_rigid integrator
    # \param group Group of particles on which to apply this method.
    # \param T Temperature of the simulation (in energy units)
    # \param seed Random seed to use for the run. Simulations that are identical, except for the seed, will follow
    # \param xi xi parameter in PSE code
    # \param error error parameter in PSE code
    # \param function_form the functional form of the strain field. If not provided, zero shear will be provided
    # \param max_strain max_strain parameter in PSE code, set that to 0.5 in most of the cases
    # \param stress_filename File name of the stress output file
    # \param stress_period The period of output of stress data
    def __init__(self, group, T, seed = 0, xi = 0.5, error = 0.001, function_form = None, max_strain = 0.5, stress_filename = None, stress_period = 100):
        util.print_status_line();

        # initialize base class
        _integration_method.__init__(self);

        # setup the variant inputs
        T = variant._setup_variant_input(T);

        # create the compute thermo
        compute._get_unique_thermo(group=group);

        self.rcut = math.sqrt( - math.log( error ) ) / xi;
        # If this line is changed, remember to change in C++ code as well!!

        # Determine whether calculate stress or not
        hydrostress_bool = False
        if stress_filename:
            hydrostress_bool = True
        else:
            stress_filename = ''

        # initialize the reflected c++ class
        if not globals.exec_conf.isCUDAEnabled():
            globals.msg.error("Sorry, we have not written CPU code for Stokesian Dynamics simulation. \n");
            raise RuntimeError('Error creating StokesRigid');
        else:
            # Create a new neighbor list
            cl_stokesrigid = hoomd.CellListGPU(globals.system_definition);
            globals.system.addCompute(cl_stokesrigid, "stokesrigid_cl")
            self.neighbor_list = hoomd.NeighborListGPUBinned(globals.system_definition, self.rcut, 0.4, cl_stokesrigid);
            self.neighbor_list.setEvery(1, True);
            globals.system.addCompute(self.neighbor_list, "stokesrigid_nlist")
            self.neighbor_list.countExclusions();

            self.cpp_method = _pse_rigidv0.StokesRigid(globals.system_definition, group.cpp_group, T.cpp_variant, seed, self.neighbor_list, xi, error, hydrostress_bool, stress_filename);

        # Set the variables
        self.cpp_method.setHydrostressPeriod(stress_period)

        self.cpp_method.validateGroup()

        if function_form is not None:
            self.cpp_method.setShear(function_form.cpp_function, max_strain)
        else:
            no_shear_function = shear_function.steady(dt = 0)
            self.cpp_method.setShear(no_shear_function.cpp_function, max_strain)

        self.cpp_method.setParams()

    ## Changes parameters of an existing integrator
    # \param T if not None: the new temperature
    # \param delimiter if not None: the new delimiter
    # \param function_form if not None: the new function form of the shear field
    # \param max_strain: 0.5 for most of the cases
    def set_params(self, T = None, delimiter = None, function_form = None, max_strain = 0.5):
        util.print_status_line();
        self.check_initialization();

        if T is not None:
            # setup the variant inputs
            T = variant._setup_variant_input(T);
            self.cpp_method.setT(T.cpp_variant);

        if delimiter:
            self.cpp_method.setDelimiter(delimiter);

        if function_form is not None:
            self.cpp_method.setShear(function_form.cpp_function, max_strain)

    ## Stop any shear
    def stop_shear(self, max_strain = 0.5):
        no_shear_function = shear_function.steady(dt = 0)
        self.cpp_method.setShear(no_shear_function.cpp_function, max_strain)
