from hoomd_script import *
from hoomd_plugins import pse_rigidv0
import unittest
import os
import numpy
import math

# unit tests for shear_function.steady
class steady_tests (unittest.TestCase):
    def setUp(self):
        init.create_random(N=1000, phi_p=0.1)

    def test_basic(self):
        steady = pse_rigidv0.shear_function.steady(dt = 1e-4)
        self.assertEqual(0.0, steady.get_shear_rate(100))
        self.assertEqual(0.0, steady.get_strain(100))
        self.assertEqual(0, steady.get_offset())
        del steady

    def test_lag(self):
        dt = 1e-4
        lj = pair.lj(r_cut = 2.5)
        lj.pair_coeff.set('A', 'A', epsilon = 1.0, sigma = 1.0)
        all = group.all()
        integrate.mode_standard(dt = dt)
        integrate.nvt(group = all, T = 1.2, tau = 0.5)
        run(10)
        steady = pse_rigidv0.shear_function.steady(dt = dt, shear_rate = 3)
        self.assertAlmostEqual(3.0, steady.get_shear_rate(20))
        self.assertAlmostEqual(3.0, steady.get_strain(10010))
        self.assertEqual(10, steady.get_offset())
        del lj, all, steady, dt

    def tearDown(self):
        init.reset()

# unit tests for shear_function.sine
class sine_tests (unittest.TestCase):
    def setUp(self):
        init.create_random(N=1000, phi_p=0.1)

    def test_basic(self):
        sine = pse_rigidv0.shear_function.sine(dt = 1e-3, shear_rate = 2.5, shear_freq = 1.0)
        self.assertAlmostEqual(2.5, sine.get_shear_rate(1000))
        self.assertAlmostEqual(0.0, sine.get_shear_rate(250))
        self.assertAlmostEqual(0.0, sine.get_strain(1000))
        self.assertAlmostEqual(0.39788736, sine.get_strain(250))
        self.assertEqual(0, sine.get_offset())
        del sine

    def test_lag(self):
        dt = 1e-4
        lj = pair.lj(r_cut = 2.5)
        lj.pair_coeff.set('A', 'A', epsilon = 1.0, sigma = 1.0)
        all = group.all()
        integrate.mode_standard(dt = dt)
        integrate.nvt(group = all, T = 1.2, tau = 0.5)
        run(500)
        sine = pse_rigidv0.shear_function.sine(dt = dt, shear_rate = 0.5, shear_freq = 10.0)
        self.assertAlmostEqual(-0.5, sine.get_shear_rate(1000))
        self.assertAlmostEqual(-0.007957747, sine.get_strain(1250))
        self.assertEqual(500, sine.get_offset())
        del lj, all, sine, dt

    def test_exception_shear(self):
        self.assertRaises(RuntimeError, pse_rigidv0.shear_function.sine, dt = 1e-3, shear_rate = 0, shear_freq = 1.0)

    def test_exception_freq(self):
        self.assertRaises(RuntimeError, pse_rigidv0.shear_function.sine, dt = 1e-3, shear_rate = 0.1, shear_freq = 0)

    def tearDown(self):
        init.reset()

# unit tests for shear_function.chirp
class chirp_tests (unittest.TestCase):
    def setUp(self):
        init.create_random(N=1000, phi_p=0.1)

    def test_basic(self):
        sine = pse_rigidv0.shear_function.chirp(dt = 1e-3, amplitude = 0.1, omega_0 = 0.1, omega_f = 1.0, periodT = 100)
        self.assertAlmostEqual(0.1 * math.cos(90/math.log(10)), sine.get_shear_rate(100000), places = 6) # places = 7 doesn't work..
        self.assertAlmostEqual(0.01 * math.sqrt(10) * math.cos((math.sqrt(10)-1) * 10 /math.log(10)), sine.get_shear_rate(50000), places = 6)
        self.assertAlmostEqual(0.01, sine.get_shear_rate(0))
        self.assertAlmostEqual(0.1 * math.sin(90/math.log(10)), sine.get_strain(100000), places = 6)
        self.assertAlmostEqual(0.1 * math.sin((math.sqrt(10)-1) * 10 /math.log(10)), sine.get_strain(50000), places = 6)
        self.assertAlmostEqual(0.0, sine.get_strain(0))
        self.assertEqual(0, sine.get_offset())
        del sine

    def test_lag(self):
        dt = 1e-4
        lj = pair.lj(r_cut = 2.5)
        lj.pair_coeff.set('A', 'A', epsilon = 1.0, sigma = 1.0)
        all = group.all()
        integrate.mode_standard(dt = dt)
        integrate.nvt(group = all, T = 1.2, tau = 0.5)
        run(500)
        sine = pse_rigidv0.shear_function.chirp(dt = dt, amplitude = 0.1, omega_0 = 0.1, omega_f = 0.5, periodT = 100)
        self.assertAlmostEqual(0.01, sine.get_shear_rate(500))
        self.assertAlmostEqual(0.05 * math.cos(40/math.log(5)), sine.get_shear_rate(1000500), places = 6)
        self.assertAlmostEqual(0.0, sine.get_strain(500))
        self.assertAlmostEqual(0.1 * math.sin(40/math.log(5)), sine.get_strain(1000500), places = 6)
        self.assertEqual(500, sine.get_offset())
        del lj, all, sine, dt

    def tearDown(self):
        init.reset()

# unit tests for shear_function.tukey_window
class tukeywindow_tests (unittest.TestCase):
    def setUp(self):
        init.create_random(N=1000, phi_p=0.1)

    def test_basic(self):
        tukey = pse_rigidv0.shear_function.tukey_window(dt = 1e-3, tukey_param = 0.5, periodT = 100)
        self.assertAlmostEqual(0.0, tukey.get_shear_rate(0), places = 6)
        self.assertAlmostEqual(0.0, tukey.get_shear_rate(25000), places = 6)
        self.assertAlmostEqual(0.0, tukey.get_shear_rate(75000), places = 6)
        self.assertAlmostEqual(1.0, tukey.get_strain(25000), places = 6)
        self.assertAlmostEqual( math.pi * math.sin(0.6*math.pi) / 50, tukey.get_shear_rate(10000), places = 6)
        self.assertAlmostEqual( (1+math.cos(-0.6*math.pi)) / 2, tukey.get_strain(10000), places = 6)
        self.assertAlmostEqual( -math.pi * math.sin(0.6*math.pi) / 50, tukey.get_shear_rate(90000), places = 6)
        self.assertAlmostEqual( (1+math.cos(0.6*math.pi)) / 2, tukey.get_strain(90000), places = 6)
        self.assertEqual(0, tukey.get_offset())
        del tukey

    def test_lag(self):
        dt = 1e-4
        lj = pair.lj(r_cut = 2.5)
        lj.pair_coeff.set('A', 'A', epsilon = 1.0, sigma = 1.0)
        all = group.all()
        integrate.mode_standard(dt = dt)
        integrate.nvt(group = all, T = 1.2, tau = 0.5)
        run(500)
        tukey = pse_rigidv0.shear_function.tukey_window(dt = dt, tukey_param = 1, periodT = 100)
        self.assertAlmostEqual(0, tukey.get_shear_rate(500))
        self.assertAlmostEqual(0, tukey.get_shear_rate(500500), places = 6)
        self.assertAlmostEqual(0, tukey.get_strain(500))
        self.assertAlmostEqual(1, tukey.get_strain(500500), places = 6)
        self.assertAlmostEqual((1 - math.cos(0.2*math.pi))/2, tukey.get_strain(100500), places = 6)
        self.assertAlmostEqual(math.sin(0.2*math.pi) * math.pi / 100, tukey.get_shear_rate(100500), places = 6)
        self.assertEqual(500, tukey.get_offset())
        del lj, all, tukey, dt

    def test_exception_negative(self):
        self.assertRaises(RuntimeError, pse_rigidv0.shear_function.tukey_window, dt = 1e-3, periodT = 100, tukey_param = 0.0)

    def test_exception_large(self):
        self.assertRaises(RuntimeError, pse_rigidv0.shear_function.tukey_window, dt = 1e-3, periodT = 100, tukey_param = 1.1)

    def tearDown(self):
        init.reset()

# unit tests for shear_function.windowed
class windowed_tests (unittest.TestCase):
    def setUp(self):
        init.create_random(N=1000, phi_p=0.1)

    def test_basic(self):
        tukey = pse_rigidv0.shear_function.tukey_window(dt = 1e-3, tukey_param = 0.1, periodT = 100)
        base = pse_rigidv0.shear_function.steady(dt = 1e-3, shear_rate = 3)
        windowed = pse_rigidv0.shear_function.windowed(function_form = base, window = tukey)
        self.assertAlmostEqual(3.0, windowed.get_shear_rate(20000), places = 6)
        self.assertAlmostEqual(0.0, windowed.get_shear_rate(200000), places = 6)
        self.assertAlmostEqual(60.0, windowed.get_strain(20000), places = 6)
        self.assertAlmostEqual(3.75, windowed.get_strain(2500), places = 6)
        self.assertAlmostEqual(0.0, windowed.get_strain(200000), places = 6)
        self.assertEqual(0, windowed.get_offset())
        del tukey

    def test_lag(self):
        dt = 1e-4
        lj = pair.lj(r_cut = 2.5)
        lj.pair_coeff.set('A', 'A', epsilon = 1.0, sigma = 1.0)
        all = group.all()
        integrate.mode_standard(dt = dt)
        integrate.nvt(group = all, T = 1.2, tau = 0.5)
        run(500)
        tukey = pse_rigidv0.shear_function.tukey_window(dt = dt, tukey_param = 1, periodT = 0.2)
        base = pse_rigidv0.shear_function.sine(dt = dt, shear_rate = 0.5, shear_freq = 10.0)
        windowed = pse_rigidv0.shear_function.windowed(function_form = base, window = tukey)
        self.assertAlmostEqual(0, windowed.get_shear_rate(500))
        self.assertAlmostEqual(-0.25, windowed.get_shear_rate(1000))
        self.assertAlmostEqual(-0.0125/math.pi * ( 1 - math.cos(0.75 * math.pi) ), windowed.get_strain(1250), places = 6)
        self.assertAlmostEqual(-math.sin(0.75 * math.pi)/8, windowed.get_shear_rate(1250), places = 6)
        self.assertEqual(500, windowed.get_offset())
        del lj, all, tukey, dt

    def tearDown(self):
        init.reset()

if __name__ == '__main__':
    unittest.main(argv = ['test.py', '-v'])
