from hoomd_script import *
from hoomd_plugins import pse_rigidv0
import unittest
import os
import numpy
import math

# unit tests for shear_variant taking in steady shear function
class steady_tests (unittest.TestCase):
    def setUp(self):
        init.create_random(N=1000, phi_p=0.1)

    def test_basic(self):
        steady = pse_rigidv0.shear_function.steady(dt = 1e-4, shear_rate = 1)
        v_steady = pse_rigidv0.variant.shear_variant(function_form = steady, total_timestep = 13000, max_strain = 1)
        self.assertAlmostEqual(0.3, v_steady.cpp_variant.getValue(3000))
        self.assertAlmostEqual(-0.7, v_steady.cpp_variant.getValue(16000))
        del steady, v_steady

    def test_lag(self):
        dt = 1e-4
        lj = pair.lj(r_cut = 2.5)
        lj.pair_coeff.set('A', 'A', epsilon = 1.0, sigma = 1.0)
        all = group.all()
        integrate.mode_standard(dt = dt)
        integrate.nvt(group = all, T = 1.2, tau = 0.5)
        run(100)
        steady = pse_rigidv0.shear_function.steady(dt = dt, shear_rate = 300)
        v_steady = pse_rigidv0.variant.shear_variant(function_form = steady, total_timestep = 310)
        self.assertAlmostEqual(0.0, v_steady.cpp_variant.getValue(0))
        self.assertAlmostEqual(0.0, v_steady.cpp_variant.getValue(100))
        self.assertAlmostEqual(-0.4, v_steady.cpp_variant.getValue(120))
        self.assertAlmostEqual(0.3, v_steady.cpp_variant.getValue(500))
        del lj, all, steady, dt, v_steady

    def tearDown(self):
        init.reset()


# unit tests for shear_variant taking in oscillatory shear function
class sine_tests (unittest.TestCase):
    def setUp(self):
        init.create_random(N=1000, phi_p=0.1)

    def test_basic(self):
        sine = pse_rigidv0.shear_function.sine(dt = 1e-2, shear_rate = 1.2, shear_freq = 1.0/2.0/math.pi)
        v_sine = pse_rigidv0.variant.shear_variant(function_form = sine, total_timestep = 1000)
        self.assertAlmostEqual(-0.14623198, v_sine.cpp_variant.getValue(235))
        self.assertAlmostEqual(-0.19999658, v_sine.cpp_variant.getValue(471))
        self.assertAlmostEqual(0.34717467, v_sine.cpp_variant.getValue(1000))
        self.assertAlmostEqual(0.34717467, v_sine.cpp_variant.getValue(2000))
        del sine, v_sine

    def tearDown(self):
        init.reset()


if __name__ == '__main__':
    unittest.main(argv = ['test.py', '-v'])
