#define BOOST_TEST_MODULE WindowFunction
#include "boost_utf_configure.h"

#include <iostream>
#include <boost/shared_ptr.hpp>

#include "ShearFunction.h"

// Test cases for TukeyWindowFunction class

// r < 1, no offset
BOOST_AUTO_TEST_CASE( tukey_basic )
{
    const boost::shared_ptr<ShearFunction> tukey( new TukeyWindowFunction(100, 0.2, 0, 1) );
    // Partition the time as:
    // 0, (0,param/2), param/2, (param/2,1-param/2), 1-param/2, (1-param/2,1), 1
    // time = 0
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getShearRate(0), tol_small);
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getStrain(0), tol_small);
    // time = param/2
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getShearRate(10), tol_small);
    MY_BOOST_CHECK_CLOSE(tukey->getStrain(10), 1, tol);
    // time = 1-param/2
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getShearRate(90), tol_small);
    MY_BOOST_CHECK_CLOSE(tukey->getStrain(90), 1, tol);
    // time = 1
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getShearRate(100), tol_small);
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getStrain(100), tol_small);
    // time in (0, param/2)
    MY_BOOST_CHECK_CLOSE(tukey->getShearRate(5), 0.15707963, tol);
    MY_BOOST_CHECK_CLOSE(tukey->getStrain(5), 0.5, tol);
    // time in (param/2, 1-param/2)
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getShearRate(35), tol_small);
    MY_BOOST_CHECK_CLOSE(tukey->getStrain(35), 1, tol);
    // time in (1-param/2,1)
    MY_BOOST_CHECK_CLOSE(tukey->getShearRate(98), -0.09232909, tol);
    MY_BOOST_CHECK_CLOSE(tukey->getStrain(98), 0.09549150, tol);

    BOOST_CHECK_EQUAL_UINT(tukey->getOffset(), 0);
}

// r = 1, with offset
BOOST_AUTO_TEST_CASE( tukey_basic_offset )
{
    const boost::shared_ptr<ShearFunction> tukey( new TukeyWindowFunction(100, 1, 37, 1) );
    // Partition the time as:
    // 0, (0,0.5), 0.5, (0.5,1), 1
    // time = 0
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getShearRate(37), tol_small);
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getStrain(37), tol_small);
    // time = param/2
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getShearRate(87), tol_small);
    MY_BOOST_CHECK_CLOSE(tukey->getStrain(87), 1, tol);
    // time = 1
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getShearRate(137), tol_small);
    MY_BOOST_CHECK_SMALL((Scalar)tukey->getStrain(137), tol_small);
    // time in (0, param/2)
    MY_BOOST_CHECK_CLOSE(tukey->getShearRate(42), 0.0097080552, tol);
    MY_BOOST_CHECK_CLOSE(tukey->getStrain(42), 0.024471742, tol);
    // time in (1-param/2,1)
    MY_BOOST_CHECK_CLOSE(tukey->getShearRate(100), -0.02290122, tol);
    MY_BOOST_CHECK_CLOSE(tukey->getStrain(100), 0.84227355, tol);

    BOOST_CHECK_EQUAL_UINT(tukey->getOffset(), 37);
}

// Test cases for WindowedFunction class

BOOST_AUTO_TEST_CASE( windowed_basic )
{
    const boost::shared_ptr<ShearFunction> steady( new SteadyShearFunction(5, 42, 0.1) );
    const boost::shared_ptr<ShearFunction> tukeyWindow( new TukeyWindowFunction(100, 0.4, 42, 0.1) );
    const boost::shared_ptr<ShearFunction> windowedSteady( new WindowedFunction(steady, tukeyWindow) );
    // Partition the time as:
    // 0, (0,param/2), param/2, (param/2,1-param/2), 1-param/2, (1-param/2,1), 1
    // time = 0
    MY_BOOST_CHECK_SMALL((Scalar)windowedSteady->getShearRate(42), tol_small);
    MY_BOOST_CHECK_SMALL((Scalar)windowedSteady->getStrain(42), tol_small);
    // time = param/2
    MY_BOOST_CHECK_CLOSE(windowedSteady->getShearRate(242), 5, tol);
    MY_BOOST_CHECK_CLOSE(windowedSteady->getStrain(242), 100, tol);
    // time = 1-param/2
    MY_BOOST_CHECK_CLOSE(windowedSteady->getShearRate(842), 5, tol);
    MY_BOOST_CHECK_CLOSE(windowedSteady->getStrain(842), 400, tol);
    // time = 1
    MY_BOOST_CHECK_SMALL((Scalar)windowedSteady->getShearRate(1042), tol_small);
    MY_BOOST_CHECK_SMALL((Scalar)windowedSteady->getStrain(1042), tol_small);
    // time in (0, param/2)
    MY_BOOST_CHECK_CLOSE(windowedSteady->getShearRate(92), 2.12063397, tol);
    MY_BOOST_CHECK_CLOSE(windowedSteady->getStrain(92), 3.66116524, tol);
    // time in (param/2, 1-param/2)
    MY_BOOST_CHECK_CLOSE(windowedSteady->getShearRate(792), 5, tol);
    MY_BOOST_CHECK_CLOSE(windowedSteady->getStrain(792), 375, tol);
    // time in (1-param/2,1)
    MY_BOOST_CHECK_CLOSE(windowedSteady->getShearRate(942), -32.84291735, tol);
    MY_BOOST_CHECK_CLOSE(windowedSteady->getStrain(942), 225, tol);

    BOOST_CHECK_EQUAL_UINT(windowedSteady->getOffset(), 42);
}
